# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import pandas as pd

from ..common.constants import ModelTask, SKLearn
from .function_wrapper import (_convert_to_two_cols, _FunctionWrapper,
                               _MultiVsSingleInstanceFunctionResolver)
from .pytorch_wrapper import WrappedPytorchModel
from .tensorflow_wrapper import WrappedTensorflowModel, is_sequential


def _is_classification_task(task):
    """Return True if the task is a classification task.

    :param task: The task to check.
    :type task: str
    :return: True if the task is a classification task.
    :rtype: bool
    """
    return task == ModelTask.CLASSIFICATION or task == ModelTask.IMAGE_CLASSIFICATION


def _eval_model(model, examples, model_task):
    """Return function from model and specify the ML Domain using model evaluation on examples.

    :param model: The model to evaluate on the examples.
    :type model: model with a predict or predict_proba function
    :param examples: The model evaluation examples.
    :type examples: ml_wrappers.DatasetWrapper
    :param model_task: Optional parameter to specify whether the model is a classification or regression model.
        In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
        has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
        outputs a 1 dimensional array.
    :type model_task: str
    :return: The function chosen from given model and chosen domain.
    :rtype: (function, str)
    """
    is_tf_seq = is_sequential(model)
    is_wrapped_pytorch = isinstance(model, WrappedPytorchModel)
    is_wrapped_tf = isinstance(model, WrappedTensorflowModel)
    if is_tf_seq or is_wrapped_pytorch or is_wrapped_tf:
        if model_task == ModelTask.REGRESSION:
            return _eval_function(model.predict, examples, ModelTask.REGRESSION)
        if model_task == ModelTask.IMAGE_CLASSIFICATION:
            examples_dataset = examples.dataset
            if isinstance(examples_dataset, pd.DataFrame):
                return _eval_function(model.predict_proba, examples,
                                      model_task, wrapped=True)
            is_pytorch_image_model = True
            wrapper = _FunctionWrapper(model.predict_proba,
                                       len(examples_dataset[0].shape),
                                       is_pytorch_image_model)
            return _eval_function(wrapper._function_input_expand_wrapper,
                                  examples, model_task, wrapped=True)
        result = model.predict_proba(examples.typed_wrapper_func(examples.dataset[0:1]))
        if result.shape[1] == 1 and model_task == ModelTask.UNKNOWN:
            raise Exception("Please specify model_task to disambiguate model type since "
                            "result of calling function is 2D array of one column.")
        else:
            return _eval_function(model.predict_proba, examples, ModelTask.CLASSIFICATION)
    else:
        has_predict_proba = hasattr(model, SKLearn.PREDICT_PROBA)
        # Note: Allow user to override default to use predict method for regressor
        if has_predict_proba and model_task != ModelTask.REGRESSION:
            return _eval_function(model.predict_proba, examples, model_task)
        else:
            return _eval_function(model.predict, examples, model_task)


def _eval_function(function, examples, model_task, wrapped=False):
    """Return function and specify the ML Domain using function evaluation on examples.

    :param function: The prediction function to evaluate on the examples.
    :type function: function
    :param examples: The model evaluation examples.
    :type examples: ml_wrappers.DatasetWrapper
    :param model_task: Optional parameter to specify whether the model is a classification or regression model.
        In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
        has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
        outputs a 1 dimensional array.
    :type model_task: str
    :param wrapped: Indicates if function has already been wrapped.
    :type wrapped: bool
    :return: The function chosen from given model and chosen domain.
    :rtype: (function, str)
    """
    # Try to run the function on a single example - if it doesn't work wrap
    # it in a function that converts a 1D array to 2D for those functions
    # that only support 2D arrays as input
    examples_dataset = examples.dataset
    if str(type(examples_dataset)).endswith(".DenseData'>"):
        examples_dataset = examples_dataset.data
    try:
        inst_result = function(examples.typed_wrapper_func(examples_dataset[0]))
        if inst_result is None:
            raise Exception("Wrapped function returned None in model wrapper when called on dataset")
        multi_inst_result = function(examples.typed_wrapper_func(examples_dataset[0:1]))
        if multi_inst_result.shape != inst_result.shape:
            if len(multi_inst_result.shape) == len(inst_result.shape) + 1:
                resolver = _MultiVsSingleInstanceFunctionResolver(function)
                return _eval_function(resolver._add_single_predict_dimension, examples, model_task)
            else:
                raise Exception("Wrapped function dimensions for single and multi predict unresolvable")
    except Exception as ex:
        # If function has already been wrapped, re-throw error to prevent stack overflow
        if wrapped:
            raise ex
        wrapper = _FunctionWrapper(function, len(examples_dataset[0].shape))
        return _eval_function(wrapper._function_input_expand_wrapper, examples, model_task, wrapped=True)
    if len(inst_result.shape) == 2:
        # If the result of evaluation the function is a 2D array of 1 column,
        # and they did not specify classifier or regressor, throw exception
        # to force the user to disambiguate the results.
        if inst_result.shape[1] == 1:
            if model_task == ModelTask.UNKNOWN:
                if isinstance(inst_result, pd.DataFrame):
                    return (function, ModelTask.REGRESSION)
                raise Exception("Please specify model_task to disambiguate model type since "
                                "result of calling function is 2D array of one column.")
            elif _is_classification_task(model_task):
                return _convert_to_two_cols(function, examples_dataset)
            else:
                # model_task == ModelTask.REGRESSION
                # In case user specified a regressor but we have a 2D output with one column,
                # we want to flatten the function to 1D
                wrapper = _FunctionWrapper(function)
                return (wrapper._function_flatten, model_task)
        else:
            if model_task == ModelTask.UNKNOWN or _is_classification_task(model_task):
                return (function, ModelTask.CLASSIFICATION)
            else:
                raise Exception("Invalid shape for prediction: "
                                "Regression function cannot output 2D array with multiple columns")
    elif len(inst_result.shape) == 1:
        if model_task == ModelTask.UNKNOWN:
            return (function, ModelTask.REGRESSION)
        elif _is_classification_task(model_task):
            return _convert_to_two_cols(function, examples_dataset)
        return (function, model_task)
    elif len(inst_result.shape) == 0:
        # single value returned, flatten output array
        wrapper = _FunctionWrapper(function)
        return (wrapper._function_flatten, model_task)
    raise Exception("Failed to wrap function, may require custom wrapper for input function or model")



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for wrapping datasets and models in one uniform format.
"""
from .dataset import DatasetWrapper
from .model import wrap_model
from .version import name, version

__all__ = ['DatasetWrapper', 'wrap_model']

import atexit
# Setup logging infrustructure
import logging
import os

# Only log to disk if environment variable specified
ml_wrappers_c_logs = os.environ.get('ML_WRAPPERS_C_LOGS')
if ml_wrappers_c_logs is not None:
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    os.makedirs(os.path.dirname(ml_wrappers_c_logs), exist_ok=True)
    handler = logging.FileHandler(ml_wrappers_c_logs, mode='w')
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.info('Initializing logging file for ml-wrappers')

    def close_handler():
        handler.close()
        logger.removeHandler(handler)
    atexit.register(close_handler)

__name__ = name
__version__ = version



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a helpful dataset wrapper to allow operations such as summarizing data, taking the subset or sampling."""

import logging
import warnings

import numpy as np
import pandas as pd
import sklearn
from packaging import version
from scipy.sparse import issparse

from ..common.constants import Defaults
from .dataset_utils import (_convert_batch_dataset_to_numpy,
                            _generate_augmented_data, _summarize_data)
from .timestamp_featurizer import CustomTimestampFeaturizer

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', 'Starting from version 2.2.1', UserWarning)


module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


try:
    import tensorflow as tf
except ImportError:
    module_logger.debug('Could not import tensorflow, required if using a Tensorflow model')
except TypeError as te:
    # handle issues with protobuf version mismatch on some environments
    module_logger.debug(
        "Could not import tensorflow in ml-wrappers due to"
        "TypeError, required if using a Tensorflow model. "
        "Inner exception: {0}".format(te))


SAMPLED_STRING_ROWS = 10


class DatasetWrapper(object):
    """A wrapper around a dataset to make dataset operations more uniform across explainers."""

    def __init__(self, dataset, clear_references=False):
        """Initialize the dataset wrapper.

        :param dataset: A matrix of feature vector examples (# examples x # features) for
            initializing the explainer.
        :type dataset: numpy.ndarray or pandas.DataFrame or panads.Series or scipy.sparse.csr_matrix
            or shap.DenseData or torch.Tensor or tensorflow.python.data.ops.dataset_ops.BatchDataset
        :param clear_references: A memory optimization that clears all references after use in explainers.
        :type clear_references: bool
        """
        if not isinstance(dataset, pd.DataFrame) and not isinstance(dataset, pd.Series) and \
                not isinstance(dataset, np.ndarray) and not issparse(dataset) and \
                not str(type(dataset)).endswith(".DenseData'>") and \
                not str(type(dataset)).endswith("torch.Tensor'>") and \
                not str(type(dataset)).endswith("BatchDataset'>"):
            raise TypeError("Got type {0} which is not supported in DatasetWrapper".format(
                type(dataset))
            )
        self._features = None
        self._original_dataset_with_type = dataset
        self._dataset_is_df = isinstance(dataset, pd.DataFrame)
        self._dataset_is_series = isinstance(dataset, pd.Series)
        self._dataset_is_batch = str(type(dataset)).endswith("BatchDataset'>")
        self._default_index_cols = ['index']
        self._default_index = True
        if self._dataset_is_df:
            self._features = dataset.columns.values.tolist()
        if self._dataset_is_df or self._dataset_is_series:
            dataset = dataset.values
        elif self._dataset_is_batch:
            dataset, features, size = _convert_batch_dataset_to_numpy(dataset)
            self._features = features
            self._batch_size = size
        self._dataset = dataset
        self._original_dataset = dataset
        self._summary_dataset = None
        self._column_indexer = None
        self._subset_taken = False
        self._summary_computed = False
        self._string_indexed = False
        self._one_hot_encoded = False
        self._one_hot_encoder = None
        self._timestamp_featurized = False
        self._timestamp_featurizer = None
        self._clear_references = clear_references

    @property
    def dataset(self):
        """Get the dataset.

        :return: The underlying dataset.
        :rtype: numpy.ndarray or scipy.sparse.csr_matrix
        """
        return self._dataset

    @property
    def typed_dataset(self):
        """Get the dataset in the original type, pandas DataFrame or Series.

        :return: The underlying dataset.
        :rtype: numpy.ndarray or pandas.DataFrame or pandas.Series or scipy.sparse matrix
        """
        wrapper_func = self.typed_wrapper_func
        return wrapper_func(self._dataset)

    def typed_wrapper_func(self, dataset, keep_index_as_feature=False):
        """Get a wrapper function to convert the dataset to the original type, pandas DataFrame or Series.

        :param dataset: The dataset to convert to original type.
        :type dataset: numpy.ndarray or scipy.sparse.csr_matrix
        :param keep_index_as_feature: Whether to keep the index as a feature when converting back.
            Off by default to convert it back to index.
        :type keep_index_as_feature: bool
        :return: A wrapper function for a given dataset to convert to original type.
        :rtype: numpy.ndarray or scipy.sparse.csr_matrix or pandas.DataFrame or pandas.Series
        """
        if self._dataset_is_df:
            if len(dataset.shape) == 1:
                dataset = dataset.reshape(1, dataset.shape[0])
            original_dtypes = self._original_dataset_with_type.dtypes
            output_types = dict(original_dtypes)
            dataframe = pd.DataFrame(dataset, columns=self._features)
            if not self._default_index:
                if keep_index_as_feature:
                    # Add the index name to type as feature dtype
                    for idx, name in enumerate(self._original_dataset_with_type.index.names):
                        level_values_dtype = self._original_dataset_with_type.index.get_level_values(idx).dtype
                        output_types.update({name: level_values_dtype})
                else:
                    dataframe = dataframe.set_index(self._default_index_cols)
            return dataframe.astype(output_types)
        elif self._dataset_is_series:
            return pd.Series(dataset)
        elif self._dataset_is_batch:
            if len(dataset.shape) == 1:
                dataset = dataset.reshape(1, dataset.shape[0])
            df = pd.DataFrame(dataset, columns=self._features)
            tensor_slices = (dict(df), None)
            tf_dataset = tf.data.Dataset.from_tensor_slices(tensor_slices)
            batch_dataset = tf_dataset.batch(self._batch_size)
            return batch_dataset
        else:
            return dataset

    @property
    def original_dataset(self):
        """Get the original dataset prior to performing any operations.

        Note: if the original dataset was a pandas dataframe, this will return the numpy version.

        :return: The original dataset.
        :rtype: numpy.ndarray or scipy.sparse matrix
        """
        return self._original_dataset

    @property
    def original_dataset_with_type(self):
        """Get the original typed dataset which could be a numpy array or pandas DataFrame or pandas Series.

        :return: The original dataset.
        :rtype: numpy.ndarray or pandas.DataFrame or pandas.Series or scipy.sparse matrix
        """
        return self._original_dataset_with_type

    @property
    def num_features(self):
        """Get the number of features (columns) on the dataset.

        :return: The number of features (columns) in the dataset.
        :rtype: int
        """
        evaluation_examples_temp = self._dataset
        if isinstance(evaluation_examples_temp, pd.DataFrame):
            evaluation_examples_temp = evaluation_examples_temp.values
        if len(evaluation_examples_temp.shape) == 1:
            return len(evaluation_examples_temp)
        elif issparse(evaluation_examples_temp):
            return evaluation_examples_temp.shape[1]
        else:
            return len(evaluation_examples_temp[0])

    @property
    def summary_dataset(self):
        """Get the summary dataset without any subsetting.

        :return: The original dataset or None if summary was not computed.
        :rtype: numpy.ndarray or scipy.sparse.csr_matrix
        """
        return self._summary_dataset

    def _set_default_index_cols(self, dataset):
        if dataset.index.names is not None:
            self._default_index_cols = dataset.index.names

    def set_index(self):
        """Undo reset_index.  Set index as feature on internal dataset to be an index again.
        """
        if self._dataset_is_df:
            dataset = self.typed_dataset
            self._features = dataset.columns.values.tolist()
            self._dataset = dataset.values
        self._default_index = True

    def reset_index(self):
        """Reset index to be part of the features on the dataset.
        """
        dataset = self._original_dataset_with_type
        if self._dataset_is_df:
            self._default_index = pd.Index(np.arange(0, len(dataset))).equals(dataset.index)
            reset_dataset = dataset
            if not self._default_index:
                self._set_default_index_cols(dataset)
                reset_dataset = dataset.reset_index()
                # Move index columns to the end of the dataframe to ensure
                # index arguments are still valid to original dataset
                dcols = reset_dataset.columns.tolist()
                for default_index_col in self._default_index_cols:
                    dcols.insert(len(dcols), dcols.pop(dcols.index(default_index_col)))
                reset_dataset = reset_dataset.reindex(columns=dcols)
            self._features = reset_dataset.columns.values.tolist()
            self._dataset = reset_dataset.values

    def get_features(self, features=None, explain_subset=None, **kwargs):
        """Get the features of the dataset if None on current kwargs.

        :return: The features of the dataset if currently None on kwargs.
        :rtype: list
        """
        if features is not None:
            if explain_subset is not None:
                return np.array(features)[explain_subset].tolist()
            return features
        if explain_subset is not None and self._features is not None:
            return np.array(self._features)[explain_subset].tolist()
        if self._features is None:
            return list(range(self._dataset.shape[1]))
        return self._features

    def get_column_indexes(self, features, categorical_features):
        """Get the column indexes for the given column names.

        :param features: The full list of existing column names.
        :type features: list[str]
        :param categorical_features: The list of categorical feature names to get indexes for.
        :type categorical_features: list[str]
        :return: The list of column indexes.
        :rtype: list[int]
        """
        return [features.index(categorical_feature) for categorical_feature in categorical_features]

    def string_index(self, columns=None):
        """Indexes categorical string features on the dataset.

        :param columns: Optional parameter specifying the subset of columns that may need to be string indexed.
        :type columns: list
        :return: The transformation steps to index the given dataset.
        :rtype: sklearn.compose.ColumnTransformer
        """
        if self._string_indexed:
            return self._column_indexer
        # Optimization so we don't redo this operation multiple times on the same dataset
        self._string_indexed = True
        # If the data was previously successfully summarized, then there are no
        # categorical columns as it must be numeric.
        # Also, if the dataset is sparse, we can assume there are no categorical strings
        if str(type(self._dataset)).endswith(".DenseData'>") or issparse(self._dataset):
            return None
        # If the user doesn't have a newer version of scikit-learn with OrdinalEncoder, don't do encoding
        try:
            from sklearn.compose import ColumnTransformer
            from sklearn.preprocessing import OrdinalEncoder
        except ImportError:
            return None
        tmp_dataset = self._dataset
        # Temporarily convert to pandas for easier and uniform string handling,
        # only use top sampled rows to limit memory usage for string type test
        if isinstance(self._dataset, np.ndarray):
            tmp_dataset = pd.DataFrame(self._dataset[:SAMPLED_STRING_ROWS, :], dtype=self._dataset.dtype)
        else:
            tmp_dataset = tmp_dataset.iloc[:SAMPLED_STRING_ROWS]
        categorical_col_names = list(np.array(list(tmp_dataset))[(tmp_dataset.applymap(type) == str).all(0)])
        if categorical_col_names:
            all_columns = tmp_dataset.columns
            if columns is not None:
                categorical_col_indices = \
                    [all_columns.get_loc(col_name) for col_name in categorical_col_names if col_name in columns]
            else:
                categorical_col_indices = [all_columns.get_loc(col_name) for col_name in categorical_col_names]
            ordinal_enc = OrdinalEncoder()
            ct = ColumnTransformer([('ord', ordinal_enc, categorical_col_indices)], remainder='drop')
            string_indexes_dataset = ct.fit_transform(self._dataset)
            # Inplace replacement of columns
            # (danger: using remainder=passthrough with ColumnTransformer will change column order!)
            for idx, categorical_col_index in enumerate(categorical_col_indices):
                self._dataset[:, categorical_col_index] = string_indexes_dataset[:, idx]
            self._column_indexer = ct
        return self._column_indexer

    def one_hot_encode(self, columns):
        """Indexes categorical string features on the dataset.

        :param columns: Parameter specifying the subset of column indexes that may need to be one-hot-encoded.
        :type columns: list[int]
        :return: The transformation steps to one-hot-encode the given dataset.
        :rtype: sklearn.preprocessing.OneHotEncoder
        """
        if self._one_hot_encoded:
            return self._one_hot_encoder
        # Optimization so we don't redo this operation multiple times on the same dataset
        self._one_hot_encoded = True
        # If the data was previously successfully summarized, then there are no
        # categorical columns as it must be numeric.
        # Also, if the dataset is sparse, we can assume there are no categorical strings
        if not columns or str(type(self._dataset)).endswith(".DenseData'>") or issparse(self._dataset):
            return None
        # If the user doesn't have a newer version of scikit-learn with OneHotEncoder, don't do encoding
        try:
            from sklearn.compose import ColumnTransformer
            from sklearn.preprocessing import OneHotEncoder
        except ImportError:
            return None
        if version.parse(sklearn.__version__) < version.parse('1.2'):
            ohe_params = {"sparse": False}
        else:
            ohe_params = {"sparse_output": False}
        one_hot_encoder = OneHotEncoder(handle_unknown='ignore', **ohe_params)
        self._one_hot_encoder = ColumnTransformer([('ord', one_hot_encoder, columns)], remainder='passthrough')
        # Note this will change column order, the one hot encoded columns will be at the start and the
        # rest of the columns at the end
        self._dataset = self._one_hot_encoder.fit_transform(self._dataset.astype(float)).astype(float)
        return self._one_hot_encoder

    def timestamp_featurizer(self):
        """Featurizes the timestamp columns.

        :return: The transformation steps to featurize the timestamp columns.
        :rtype: ml_wrappers.DatasetWrapper
        """
        if self._timestamp_featurized:
            return self._timestamp_featurizer
        # Optimization so we don't redo this operation multiple times on the same dataset
        self._timestamp_featurized = True
        # If the data was previously successfully summarized, then there are no
        # categorical columns as it must be numeric.
        # Also, if the dataset is sparse, we can assume there are no categorical strings
        if str(type(self._dataset)).endswith(".DenseData'>") or issparse(self._dataset):
            return None
        typed_dataset_without_index = self.typed_wrapper_func(self._dataset, keep_index_as_feature=True)
        self._timestamp_featurizer = CustomTimestampFeaturizer(self._features).fit(typed_dataset_without_index)
        self._dataset = self._timestamp_featurizer.transform(self._dataset)
        return self._timestamp_featurizer

    def apply_indexer(self, column_indexer, bucket_unknown=False):
        """Indexes categorical string features on the dataset.

        :param column_indexer: The transformation steps to index the given dataset.
        :type column_indexer: sklearn.compose.ColumnTransformer
        :param bucket_unknown: If true, buckets unknown values to separate categorical level.
        :type bucket_unknown: bool
        """
        if self._string_indexed or issparse(self._dataset):
            return
        name, ordinal_encoder, cols = column_indexer.transformers_[0]
        all_categories = ordinal_encoder.categories_

        def convert_cols(category_to_index, value, unknown):
            if value in category_to_index:
                index = category_to_index[value]
            elif not bucket_unknown:
                # Add new index on the fly - note the background data does NOT need to
                # contain all possible category levels!
                index = len(category_to_index) + 1
                category_to_index[value] = index
            else:
                # Put all unknown indexes into a separate 'unknown' bucket
                index = unknown
                category_to_index[value] = index
            return index

        for idx, i in enumerate(cols):
            categories_for_col = all_categories[idx]
            category_to_index = dict(zip(categories_for_col, range(len(categories_for_col))))
            unknown = len(category_to_index) + 1
            self._dataset[:, i] = list(map(lambda x: convert_cols(category_to_index, x, unknown), self._dataset[:, i]))
        # Ensure element types are float and not object
        self._dataset = self._dataset.astype(float)
        self._string_indexed = True

    def apply_one_hot_encoder(self, one_hot_encoder):
        """One-hot-encode categorical string features on the dataset.

        :param one_hot_encoder: The transformation steps to one-hot-encode the given dataset.
        :type one_hot_encoder: sklearn.preprocessing.OneHotEncoder
        """
        if self._one_hot_encoded or issparse(self._dataset):
            return
        self._dataset = one_hot_encoder.transform(self._dataset).astype(float)
        self._one_hot_encoded = True

    def apply_timestamp_featurizer(self, timestamp_featurizer):
        """Apply timestamp featurization on the dataset.

        :param timestamp_featurizer: The transformation steps to featurize timestamps in the given dataset.
        :type timestamp_featurizer: CustomTimestampFeaturizer
        """
        if self._timestamp_featurized or issparse(self._dataset):
            return
        self._dataset = timestamp_featurizer.transform(self._dataset)
        self._timestamp_featurized = True

    def compute_summary(self, nclusters=10, use_gpu=False, **kwargs):
        """Summarizes the dataset if it hasn't been summarized yet."""
        if self._summary_computed:
            return
        self._summary_dataset = _summarize_data(self._dataset, nclusters, use_gpu)
        self._dataset = self._summary_dataset
        self._summary_computed = True

    def augment_data(self, max_num_of_augmentations=np.inf):
        """Augment the current dataset.

        :param max_augment_data_size: number of times we stack permuted x to augment.
        :type max_augment_data_size: int
        """
        self._dataset = _generate_augmented_data(self._dataset, max_num_of_augmentations=max_num_of_augmentations)

    def take_subset(self, explain_subset):
        """Take a subset of the dataset if not done before.

        :param explain_subset: A list of column indexes to take from the original dataset.
        :type explain_subset: list
        """
        if self._subset_taken:
            return
        # Edge case: Take the subset of the summary in this case,
        # more optimal than recomputing the summary!
        explain_subset = np.array(explain_subset)
        self._dataset = self._dataset[:, explain_subset]
        self._subset_taken = True

    def _reduce_examples(self, max_dim_clustering=Defaults.MAX_DIM):
        """Reduces the dimensionality of the examples if dimensionality is higher than max_dim_clustering.

        If the dataset is sparse, we mean-scale the data and then run
        truncated SVD to reduce the number of features to max_dim_clustering.  For dense
        dataset, we also scale the data and then run PCA to reduce the number of features to
        max_dim_clustering.
        This is used to get better clustering results in _find_k.

        :param max_dim_clustering: Dimensionality threshold for performing reduction.
        :type max_dim_clustering: int
        """
        from sklearn.decomposition import PCA, TruncatedSVD
        from sklearn.preprocessing import StandardScaler
        num_cols = self._dataset.shape[1]
        # Run PCA or SVD on input data and reduce to about MAX_DIM features prior to clustering
        components = min(max_dim_clustering, num_cols)
        reduced_examples = self._dataset
        if components != num_cols:
            if issparse(self._dataset):
                module_logger.debug('Reducing sparse data with StandardScaler and TruncatedSVD')
                normalized_examples = StandardScaler(with_mean=False).fit_transform(self._dataset)
                reducer = TruncatedSVD(n_components=components)
            else:
                module_logger.debug('Reducing normal data with StandardScaler and PCA')
                normalized_examples = StandardScaler().fit_transform(self._dataset)
                reducer = PCA(n_components=components)
            module_logger.info('reducing dimensionality to {0} components for clustering'.format(str(components)))
            reduced_examples = reducer.fit_transform(normalized_examples)
        return reduced_examples

    def _find_k_kmeans(self, max_dim_clustering=Defaults.MAX_DIM):
        """Use k-means to downsample the examples.

        Starting from k_upper_bound, cuts k in half each time and run k-means
        clustering on the examples.  After each run, computes the
        silhouette score and stores k with highest silhouette score.
        We use optimal k to determine how much to downsample the examples.

        :param max_dim_clustering: Dimensionality threshold for performing reduction.
        :type max_dim_clustering: int
        """
        from math import ceil, isnan, log

        from sklearn.cluster import KMeans
        from sklearn.metrics import silhouette_score
        reduced_examples = self._reduce_examples(max_dim_clustering)
        num_rows = self._dataset.shape[0]
        k_upper_bound = 2000
        k_list = []
        k = min(num_rows / 2, k_upper_bound)
        for _ in range(int(ceil(log(num_rows, 2) - 7))):
            k_list.append(int(k))
            k /= 2
        prev_highest_score = -1
        prev_highest_index = 0
        opt_k = int(k)
        for k_index, k in enumerate(k_list):
            module_logger.info('running KMeans with k: {}'.format(str(k)))
            km = KMeans(n_clusters=k).fit(reduced_examples)
            clusters = km.labels_
            num_clusters = len(set(clusters))
            k_too_big = num_clusters <= 1
            if k_too_big or num_clusters == reduced_examples.shape[0]:
                score = -1
            else:
                score = silhouette_score(reduced_examples, clusters)
            if isnan(score):
                score = -1
            module_logger.info('KMeans silhouette score: {}'.format(str(score)))
            # Find k with highest silhouette score for optimal clustering
            if score >= prev_highest_score and not k_too_big:
                prev_highest_score = score
                prev_highest_index = k_index
        opt_k = k_list[prev_highest_index]
        module_logger.info('best silhouette score: {}'.format(str(prev_highest_score)))
        module_logger.info('optimal k for KMeans: {}'.format(str(opt_k)))
        return opt_k

    def _find_k_hdbscan(self, max_dim_clustering=Defaults.MAX_DIM):
        """Use hdbscan to downsample the examples.

        We use optimal k to determine how much to downsample the examples.

        :param max_dim_clustering: Dimensionality threshold for performing reduction.
        :type max_dim_clustering: int
        """
        import hdbscan
        num_rows = self._dataset.shape[0]
        reduced_examples = self._reduce_examples(max_dim_clustering)
        hdbs = hdbscan.HDBSCAN(min_cluster_size=2).fit(reduced_examples)
        clusters = hdbs.labels_
        opt_k = len(set(clusters))
        clustering_threshold = 5
        samples = opt_k * clustering_threshold
        module_logger.info(('found optimal k for hdbscan: {},'
                            ' will use clustering_threshold * k for sampling: {}').format(str(opt_k), str(samples)))
        return min(samples, num_rows)

    def sample(self, max_dim_clustering=Defaults.MAX_DIM, sampling_method=Defaults.HDBSCAN):
        """Sample the examples.

        First does random downsampling to upper_bound rows,
        then tries to find the optimal downsample based on how many clusters can be constructed
        from the data.  If sampling_method is hdbscan, uses hdbscan to cluster the
        data and then downsamples to that number of clusters.  If sampling_method is k-means,
        uses different values of k, cutting in half each time, and chooses the k with highest
        silhouette score to determine how much to downsample the data.
        The danger of using only random downsampling is that we might downsample too much
        or too little, so the clustering approach is a heuristic to give us some idea of
        how much we should downsample to.

        :param max_dim_clustering: Dimensionality threshold for performing reduction.
        :type max_dim_clustering: int
        :param sampling_method: Method to use for sampling, can be 'hdbscan' or 'kmeans'.
        :type sampling_method: str
        """
        from sklearn.utils import resample

        # bounds are rough estimates that came from manual investigation
        lower_bound = 200
        upper_bound = 10000
        num_rows = self._dataset.shape[0]
        module_logger.info('sampling examples')
        # If less than lower_bound rows, just return the full dataset
        if num_rows < lower_bound:
            return self._dataset
        # If more than upper_bound rows, sample randomly
        elif num_rows > upper_bound:
            module_logger.info('randomly sampling to 10k rows')
            self._dataset = resample(self._dataset, n_samples=upper_bound, random_state=7)
            num_rows = upper_bound
        if sampling_method == Defaults.HDBSCAN:
            try:
                opt_k = self._find_k_hdbscan(max_dim_clustering)
            except Exception as ex:
                module_logger.warning(('Failed to use hdbscan due to error: {}'
                                      '\nEnsure hdbscan is installed with: pip install hdbscan').format(str(ex)))
                opt_k = self._find_k_kmeans(max_dim_clustering)
        else:
            opt_k = self._find_k_kmeans(max_dim_clustering)
        # Resample based on optimal number of clusters
        if (opt_k < num_rows):
            self._dataset = resample(self._dataset, n_samples=opt_k, random_state=7)
        return self._dataset

    def _clear(self):
        """Optimization for memory usage.

        Clears all internal references so they can be garbage collected.
        """
        if self._clear_references:
            self._features = None
            self._original_dataset_with_type = None
            self._dataset_is_df = None
            self._dataset_is_series = None
            self._default_index_cols = None
            self._default_index = None
            self._dataset = None
            self._original_dataset = None
            self._summary_dataset = None
            self._column_indexer = None
            self._subset_taken = False
            self._summary_computed = False
            self._string_indexed = False
            self._one_hot_encoded = False
            self._one_hot_encoder = None
            self._timestamp_featurized = False
            self._timestamp_featurizer = None



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a model wrapper for an openai model endpoint."""

import numpy as np
import pandas as pd

try:
    import openai
    openai_installed = True
except ImportError:
    openai_installed = False
try:
    from openai import AzureOpenAI, OpenAI
except ImportError:
    # Ignore the error, only used by new openai version
    pass
try:
    from raiutils.common.retries import retry_function
    rai_utils_installed = True
except ImportError:
    rai_utils_installed = False


AZURE = 'azure'
CHAT_COMPLETION = 'ChatCompletion'
CONTENT = 'content'
OPENAI = 'OpenAI'
HISTORY = 'history'
SYS_PROMPT = 'sys_prompt'


def replace_backtick_chars(message):
    """Replace backtick characters in a message.

    :param message: The message.
    :type message: str
    :return: The message with backtick characters replaced.
    :rtype: str
    """
    return message.replace('`', '')


class ChatCompletion(object):
    """A class to call the openai chat completion endpoint."""

    def __init__(self, messages, engine, temperature,
                 max_tokens, top_p, frequency_penalty,
                 presence_penalty, stop, client=None):
        """Initialize the class.

        :param messages: The messages.
        :type messages: list
        :param engine: The engine.
        :type engine: str
        :param temperature: The temperature.
        :type temperature: float
        :param max_tokens: The maximum number of tokens.
        :type max_tokens: int
        :param top_p: The top p.
        :type top_p: float
        :param frequency_penalty: The frequency penalty.
        :type frequency_penalty: float
        :param presence_penalty: The presence penalty.
        :type presence_penalty: float
        :param stop: The stop.
        :type stop: list
        :param client: The client, if using openai>1.0.0.
        :type client: openai.OpenAI
        """
        self.messages = messages
        self.engine = engine
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.top_p = top_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.stop = stop
        self.client = client

    def fetch(self):
        """Call the openai chat completion endpoint.

        :return: The response.
        :rtype: dict
        """
        if not hasattr(openai, OPENAI):
            # openai<1.0.0
            return openai.ChatCompletion.create(
                engine=self.engine,
                messages=self.messages,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stop=self.stop)

        else:
            # openai>=1.0.0
            return self.client.chat.completions.create(
                model=self.engine,
                messages=self.messages,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stop=self.stop)


class OpenaiWrapperModel(object):
    """A model wrapper for an openai model endpoint."""

    def __init__(self, api_type, api_base, api_version, api_key,
                 engine="gpt-4-32k", temperature=0.7, max_tokens=800,
                 top_p=0.95, frequency_penalty=0, presence_penalty=0,
                 stop=None, input_col='prompt'):
        """Initialize the model.

        :param api_type: The type of the API.
        :type api_type: str
        :param api_base: The base URL for the API.
        :type api_base: str
        :param api_version: The version of the API.
        :type api_version: str
        :param api_key: The API key.
        :type api_key: str
        :param engine: The engine.
        :type engine: str
        :param temperature: The temperature.
        :type temperature: float
        :param max_tokens: The maximum number of tokens.
        :type max_tokens: int
        :param top_p: The top p.
        :type top_p: float
        :param frequency_penalty: The frequency penalty.
        :type frequency_penalty: float
        :param presence_penalty: The presence penalty.
        :type presence_penalty: float
        :param stop: The stop.
        :type stop: list
        :param input_col: The name of the input column.
        :type input_col: str
        """
        self.api_type = api_type
        self.api_base = api_base
        self.api_version = api_version
        self.api_key = api_key
        self.engine = engine
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.top_p = top_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.stop = stop
        self.input_col = input_col

    def _call_webservice(self, data, history=None, sys_prompt=None):
        """Common code to call the webservice.

        :param data: The data to send to the webservice.
        :type data: pandas.Series
        :param history: The history.
        :type history: pandas.Series
        :param sys_prompt: The system prompt.
        :type sys_prompt: pandas.Series
        :return: The result.
        :rtype: numpy.ndarray
        """
        if not rai_utils_installed:
            error = "raiutils package is required to call openai endpoint"
            raise RuntimeError(error)
        if not openai_installed:
            error = "openai package is required to call openai endpoint"
            raise RuntimeError(error)
        if not isinstance(data, list):
            if isinstance(data, np.ndarray):
                data = data.tolist()
            else:
                data = data.values.tolist()
        client = None
        if hasattr(openai, OPENAI):
            if self.api_type == AZURE:
                client = AzureOpenAI(api_key=self.api_key, azure_endpoint=self.api_base,
                                     api_version=self.api_version)
            else:
                client = OpenAI(api_key=self.api_key)
        else:
            openai.api_key = self.api_key
            openai.api_base = self.api_base
            openai.api_type = self.api_type
            openai.api_version = self.api_version
        answers = []
        for i, doc in enumerate(data):
            messages = []
            if sys_prompt is not None:
                messages.append({'role': 'system', CONTENT: sys_prompt.iloc[i]})
            if history is not None:
                messages.extend(history.iloc[i])
            messages.append({'role': 'user', CONTENT: doc})
            fetcher = ChatCompletion(messages, self.engine, self.temperature,
                                     self.max_tokens, self.top_p,
                                     self.frequency_penalty,
                                     self.presence_penalty, self.stop, client)
            action_name = "Call openai chat completion"
            err_msg = "Failed to call openai endpoint"
            max_retries = 4
            retry_delay = 60
            response = retry_function(fetcher.fetch, action_name, err_msg,
                                      max_retries=max_retries,
                                      retry_delay=retry_delay)
            if isinstance(response, dict):
                answers.append(replace_backtick_chars(response['choices'][0]['message'][CONTENT]))
            else:
                answers.append(replace_backtick_chars(response.choices[0].message.content))
        return np.array(answers)

    def _get_input_data(self, model_input, input_col):
        if isinstance(model_input, dict):
            prompts = pd.Series(model_input[input_col])
            if HISTORY in model_input:
                if isinstance(model_input[input_col], str):
                    history = pd.Series([model_input[HISTORY]])
                else:
                    history = pd.Series(model_input[HISTORY])
            else:
                history = None
            if SYS_PROMPT in model_input:
                sys_prompt = pd.Series(model_input[SYS_PROMPT])
            else:
                sys_prompt = None
        else:
            prompts = model_input[input_col]
            history = model_input.get(HISTORY)
            sys_prompt = model_input.get(SYS_PROMPT)

        return prompts, history, sys_prompt

    def predict(self, context, model_input=None):
        """Predict using the model.

        :param context: The context for MLFlow model or the input data.
        :type context: mlflow.pyfunc.model.PythonModelContext or
            pandas.DataFrame
        :param model_input: The input to the model.
        :type model_input: pandas.DataFrame or dict or list[str]
            pandas.Series or str
        :return: The predictions.
        :rtype: numpy.ndarray
        """
        # This is to conform to the scikit-learn API format
        # which MLFlow does not follow
        if model_input is None:
            model_input = context

        if isinstance(model_input, str):
            model_input = [model_input]
        if isinstance(model_input, (list, pd.Series)):
            questions = pd.Series(model_input)
            history = None
            sys_prompt = None
        else:
            try:
                questions, history, sys_prompt = self._get_input_data(model_input, self.input_col)
            except KeyError:
                # Fallback option keep support for older versions
                questions, history, sys_prompt = self._get_input_data(model_input, 'questions')

        result = self._call_webservice(
            questions,
            history=history,
            sys_prompt=sys_prompt)
        return result



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines common model utilities."""

from ml_wrappers.common.warnings_suppressor import shap_warnings_suppressor

with shap_warnings_suppressor():
    try:
        from shap.utils import safe_isinstance
        shap_installed = True
    except BaseException:  # noqa: B036
        shap_installed = False


MULTILABEL_THRESHOLD = 0.5


def _is_transformers_pipeline(model):
    """Checks if the model is a transformers pipeline.

    :param model: The model to check.
    :type model: object
    :return: True if the model is a transformers pipeline, False otherwise.
    :rtype: bool
    """
    return shap_installed and safe_isinstance(
        model, "transformers.pipelines.Pipeline")


def _is_callable_pipeline(model):
    """Checks if the model is a callable pipeline.

    Returns false if the model has a predict and predict_proba method.

    :param model: The model to check.
    :type model: object
    :return: True if the model is a callable pipeline, False otherwise.
    :rtype: bool
    """
    has_predict = hasattr(model, 'predict')
    has_predict_proba = hasattr(model, 'predict_proba')
    return callable(model) and not has_predict and not has_predict_proba



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Common infrastructure, class hierarchy and utilities for model explanations."""

from .endpoint_wrapper import EndpointWrapperModel
from .model_wrapper import _wrap_model, wrap_model
from .openai_wrapper import OpenaiWrapperModel
from .pytorch_wrapper import WrappedPytorchModel
from .tensorflow_wrapper import WrappedTensorflowModel, is_sequential
from .wrapped_classification_model import WrappedClassificationModel
from .wrapped_regression_model import WrappedRegressionModel

__all__ = ['EndpointWrapperModel', 'OpenaiWrapperModel',
           'WrappedClassificationModel', 'WrappedPytorchModel',
           'WrappedRegressionModel', 'WrappedTensorflowModel',
           '_wrap_model', 'is_sequential', 'wrap_model']



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a class for wrapping classification models."""

import numpy as np
import pandas as pd
from ml_wrappers.common.constants import ModelTask, SKLearn
from ml_wrappers.model.base_wrapped_model import BaseWrappedModel
from ml_wrappers.model.function_wrapper import _FunctionWrapper
from ml_wrappers.model.pytorch_wrapper import WrappedPytorchModel
from ml_wrappers.model.tensorflow_wrapper import (WrappedTensorflowModel,
                                                  is_sequential)


class WrappedClassificationModel(BaseWrappedModel):
    """A class for wrapping a classification model."""

    def __init__(self, model, eval_function, examples=None):
        """Initialize the WrappedClassificationModel with the model and evaluation function."""
        super(WrappedClassificationModel, self).__init__(model, eval_function, examples, ModelTask.CLASSIFICATION)

    def predict(self, dataset):
        """Predict the output using the wrapped classification model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        is_tf_seq = is_sequential(self._model)
        is_wrapped_pytroch = isinstance(self._model, WrappedPytorchModel)
        is_wrapped_tf = isinstance(self._model, WrappedTensorflowModel)
        if is_tf_seq or is_wrapped_pytroch or is_wrapped_tf:
            wrapped_predict_classes = self._wrap_function(self._model.predict_classes)
            return wrapped_predict_classes(dataset).flatten()
        wrapped_predict = self._wrap_function(self._model.predict)
        preds = wrapped_predict(dataset)
        if isinstance(preds, pd.DataFrame):
            preds = preds.values.ravel()
        # Handle possible case where the model has only a predict function and it outputs probabilities
        # Note this is different from WrappedClassificationWithoutProbaModel where there is no predict_proba
        # method but the predict method outputs classes
        has_predict_proba = hasattr(self._model, SKLearn.PREDICT_PROBA)
        if not has_predict_proba:
            if len(preds.shape) == 1:
                return np.argmax(preds)
            else:
                return np.argmax(preds, axis=1)
        # Handle the case that the model predicts a two-dimensional array of one column
        if len(preds.shape) == 2 and preds.shape[1] == 1:
            preds = preds.ravel()
        return np.array(preds)

    def predict_proba(self, dataset):
        """Predict the output probability using the wrapped model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        proba_preds = self._eval_function(dataset)
        if isinstance(proba_preds, pd.DataFrame):
            proba_preds = proba_preds.values

        return proba_preds

    def _wrap_function(self, function):
        """Wrap a function to conform to the prediction input contracts.

        If model requires _function_input_expand_wrapper, re-wraps
        the given function with _function_input_expand_wrapper.

        :param function: The function to wrap.
        :type function: function
        :return: The wrapped function.
        :rtype: function
        """
        eval_function = self._eval_function
        exp_wrapper = _FunctionWrapper._function_input_expand_wrapper
        exp_wrapper_name = exp_wrapper.__name__
        if eval_function.__name__ == exp_wrapper_name:
            base_dims = eval_function.__self__._base_dims
            function_wrapper = _FunctionWrapper(function, base_dims)
            function = function_wrapper._function_input_expand_wrapper
        return function



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines model wrappers and utilities for fastai tabular models."""

import numpy as np

FASTAI_TABULAR_MODEL_SUFFIX = "fastai.tabular.learner.TabularLearner'>"


def _is_fastai_tabular_model(model):
    """Check if the model is a fastai tabular model.

    :param model: The model to check.
    :type model: object
    :return: True if the model is a fastai model, False otherwise.
    :rtype: bool
    """
    return str(type(model)).endswith(FASTAI_TABULAR_MODEL_SUFFIX)


class WrappedFastAITabularModel(object):
    """A class for wrapping a FastAI tabular model in the scikit-learn style."""

    def __init__(self, model):
        """Initialize the WrappedFastAITabularModel.

        :param model: The model to wrap.
        :type model: fastai.learner.Learner
        """
        self._model = model
        dl = self._model.dls[0]
        self.cat_cols = dl.dataset.cat_names
        self.cont_cols = dl.dataset.cont_names

    def _fastai_predict(self, dataset, index, model=None):
        """Predict the output using the wrapped FastAI model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :param index: The index into the predicted data.
            Index 1 is for the predicted class and index
            2 is for the predicted probability.
        :type index: int
        :param model: The model to use for prediction.
            If None, the wrapped model is used.
        :type model: fastai.learner.Learner
        :return: The predicted data.
        :rtype: numpy.ndarray
        """
        if model is None:
            model = self._model
        predictions = []
        for i in range(len(dataset)):
            row = dataset.iloc[i]
            # get only feature columns for prediction
            row = row[self.cat_cols + self.cont_cols]
            predictions.append(np.array(model.predict(row)[index]))
        predictions = np.array(predictions)
        if index == 1:
            is_boolean = predictions.dtype == bool
            if is_boolean:
                predictions = predictions.astype(int)
        return predictions

    def _fastai_predict_without_callbacks(self, dataset, index):
        """Predict the output using the wrapped FastAI model without callbacks.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :param index: The index into the predicted data.
            Index 1 is for the predicted class and index
            2 is for the predicted probability.
        :type index: int
        :return: The predicted data.
        :rtype: numpy.ndarray
        """
        removed_cbs = []
        default_cbs = ['TrainEvalCallback', 'Recorder', 'CastToTensor']
        for cb in self._model.cbs:
            if cb.__class__.__name__ not in default_cbs:
                removed_cbs.append(cb)
        with self._model.removed_cbs(removed_cbs) as model:
            return self._fastai_predict(dataset, index, model)

    def predict(self, dataset):
        """Predict the output value using the wrapped FastAI model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted values.
        :rtype: numpy.ndarray
        """
        try:
            return self._fastai_predict(dataset, 1)
        except Exception:
            return self._fastai_predict_without_callbacks(dataset, 1)

    def predict_proba(self, dataset):
        """Predict the output probability using the FastAI model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted probabilities.
        :rtype: numpy.ndarray
        """
        try:
            return self._fastai_predict(dataset, 2)
        except Exception:
            return self._fastai_predict_without_callbacks(dataset, 2)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a base class for wrapping models."""

from ml_wrappers.model.evaluator import _eval_model


class BaseWrappedModel(object):
    """A base class for WrappedClassificationModel and WrappedRegressionModel."""

    def __init__(self, model, eval_function, examples, model_task):
        """Initialize the WrappedClassificationModel with the model and evaluation function."""
        self._eval_function = eval_function
        self._model = model
        self._examples = examples
        self._model_task = model_task

    def __getstate__(self):
        """Influence how BaseWrappedModel is pickled.

        Removes _eval_function which may not be serializable.

        :return state: The state to be pickled, with _eval_function removed.
        :rtype: dict
        """
        odict = self.__dict__.copy()
        if self._examples is not None:
            del odict['_eval_function']
        return odict

    def __setstate__(self, state):
        """Influence how BaseWrappedModel is unpickled.

        Re-adds _eval_function which may not be serializable.

        :param dict: A dictionary of deserialized state.
        :type dict: dict
        """
        self.__dict__.update(state)
        if self._examples is not None:
            eval_function, _ = _eval_model(self._model, self._examples, self._model_task)
            self._eval_function = eval_function



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines wrappers for text-based models."""

import numpy as np
from ml_wrappers.common.constants import ModelTask
from ml_wrappers.common.warnings_suppressor import shap_warnings_suppressor
from ml_wrappers.model.model_utils import (MULTILABEL_THRESHOLD,
                                           _is_transformers_pipeline)

with shap_warnings_suppressor():
    try:
        from shap import models
        shap_installed = True
    except BaseException:  # noqa: B036
        shap_installed = False


def _wrap_text_model(model, examples, model_task, is_function):
    """If needed, wraps the model or function in a common API.

    Wraps the model based on model task and prediction function contract.

    :param model: The model or function to evaluate on the examples.
    :type model: function or model with a predict or predict_proba function
    :param examples: The model evaluation examples.
        Note the examples will be wrapped in a DatasetWrapper, if not
        wrapped when input.
    :type examples: ml_wrappers.DatasetWrapper or numpy.ndarray
        or pandas.DataFrame or panads.Series or scipy.sparse.csr_matrix
        or shap.DenseData or torch.Tensor
    :param model_task: Parameter to specify whether the model is a
        'text_classification', 'sentiment_analysis', 'question_answering',
        'entailment' or 'summarizations' text model.
    :type model_task: str
    :return: The function chosen from given model and chosen domain, or
        model wrapping the function and chosen domain.
    :rtype: (function, str) or (model, str)
    """
    _wrapped_model = model
    if _is_transformers_pipeline(model):
        if model_task == ModelTask.TEXT_CLASSIFICATION:
            _wrapped_model = WrappedTextClassificationModel(model)
        elif model_task == ModelTask.QUESTION_ANSWERING:
            _wrapped_model = WrappedQuestionAnsweringModel(model)
        elif model_task == ModelTask.MULTILABEL_TEXT_CLASSIFICATION:
            _wrapped_model = WrappedTextClassificationModel(model, multilabel=True)
    return _wrapped_model, model_task


class WrappedTextClassificationModel(object):
    """A class for wrapping a Transformers model in the scikit-learn style."""

    def __init__(self, model, multilabel=False):
        """Initialize the WrappedTextClassificationModel."""
        self._model = model
        if not shap_installed:
            raise ImportError("SHAP is not installed. Please install it "
                              "to use WrappedTextClassificationModel.")
        self._wrapped_model = models.TransformersPipeline(model)
        self._multilabel = multilabel

    def predict(self, dataset):
        """Predict the output using the wrapped Transformers model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        pipeline_dicts = self._wrapped_model.inner_model(dataset)
        output = []
        for val in pipeline_dicts:
            if not isinstance(val, list):
                val = [val]
            scores = [obj["score"] for obj in val]
            if self._multilabel:
                threshold = MULTILABEL_THRESHOLD
                # jagged, thresholded array of labels model predicted
                labels = np.where(np.array(scores) > threshold)
                predictions = np.zeros(len(scores))
                # indicator matrix of labels since numpy does not
                # support jagged arrays, which seems to be the format
                # scikit-learn MultiOutputClassifier uses,
                # see sklearn.multioutput.MultiOutputClassifier.predict
                predictions[labels] = 1
                output.append(predictions)
            else:
                max_score_index = np.argmax(scores)
                output.append(max_score_index)
        return np.array(output)

    def predict_proba(self, dataset):
        """Predict the output probability using the Transformers model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        return self._wrapped_model(dataset)


class WrappedQuestionAnsweringModel(object):
    """A class for wrapping a Transformers model in the scikit-learn style."""

    def __init__(self, model):
        """Initialize the WrappedQuestionAnsweringModel."""
        self._model = model

    def predict(self, dataset):
        """Predict the output using the wrapped Transformers model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        output = []
        for context, question in zip(dataset['context'], dataset['questions']):
            answer = self._model({'context': context, 'question': question})
            output.append(answer['answer'])
        return np.array(output)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a custom timestamp featurizer for converting timestamp columns to numeric."""

import numpy as np
import pandas as pd
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from scipy.sparse import issparse
from sklearn.base import BaseEstimator, TransformerMixin


class CustomTimestampFeaturizer(BaseEstimator, TransformerMixin):
    """An estimator for featurizing timestamp columns to numeric data.

    :param features: Optional feature column names.
    :type features: list[str]
    :param return_pandas: Whether to return the transformed dataset as a pandas DataFrame.
    :type return_pandas: bool
    :param modify_in_place: Whether to modify the original dataset in place.
    :type modify_in_place: bool
    """

    def __init__(self, features=None, return_pandas=False, modify_in_place=False):
        """Initialize the CustomTimestampFeaturizer.

        :param features: Optional feature column names.
        :type features: list[str]
        :param return_pandas: Whether to return the transformed dataset as a pandas DataFrame.
        :type return_pandas: bool
        :param modify_in_place: Whether to modify the original dataset in place.
        :type modify_in_place: bool
        """
        self.features = features
        self.return_pandas = return_pandas
        self.modify_in_place = modify_in_place
        self._time_col_names = []
        return

    def fit(self, X, y=None):
        """Fits the CustomTimestampFeaturizer.

        :param X: The dataset containing timestamp columns to featurize.
        :type X: numpy.ndarray or pandas.DataFrame or scipy.sparse.csr_matrix
        :param y: The target values.
        :type y: Optional target values (None for unsupervised transformations).
        """
        # If the data was previously successfully summarized, then there are no
        # timestamp columns as it must be numeric.
        # Also, if the dataset is sparse, we can assume there are no timestamps
        if str(type(X)).endswith(".DenseData'>") or issparse(X):
            return self
        tmp_dataset = X
        # If numpy array, temporarily convert to pandas for easier and uniform timestamp handling
        if isinstance(X, np.ndarray):
            tmp_dataset = pd.DataFrame(X, columns=self.features)
        self._time_col_names = [column for column in tmp_dataset.columns if is_datetime(tmp_dataset[column])]
        # Calculate the min date for each column
        self._min = []
        for time_col_name in self._time_col_names:
            self._min.append(tmp_dataset[time_col_name].map(lambda x: x.timestamp()).min())
        return self

    def transform(self, X):
        """Transforms the timestamp columns to numeric type in the given dataset.

        Specifically, extracts the year, month, day, hour, minute, second and time
        since min timestamp in the training dataset.

        :param X: The dataset containing timestamp columns to featurize.
        :type X: numpy.ndarray or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The transformed dataset.
        :rtype: numpy.ndarray or scipy.sparse.csr_matrix
        """
        tmp_dataset = X
        if len(self._time_col_names) > 0:
            # Temporarily convert to pandas for easier and uniform timestamp handling
            if isinstance(X, np.ndarray):
                tmp_dataset = pd.DataFrame(X, columns=self.features)
            elif not self.modify_in_place:
                # If originally pandas, make a copy to avoid changing the original dataset
                tmp_dataset = X.copy()
            # Get the year, day, month, hour, minute, second
            for idx, time_col_name in enumerate(self._time_col_names):
                tmp_dataset[time_col_name + '_year'] = tmp_dataset[time_col_name].map(lambda x: x.year)
                tmp_dataset[time_col_name + '_month'] = tmp_dataset[time_col_name].map(lambda x: x.month)
                tmp_dataset[time_col_name + '_day'] = tmp_dataset[time_col_name].map(lambda x: x.day)
                tmp_dataset[time_col_name + '_hour'] = tmp_dataset[time_col_name].map(lambda x: x.hour)
                tmp_dataset[time_col_name + '_minute'] = tmp_dataset[time_col_name].map(lambda x: x.minute)
                tmp_dataset[time_col_name + '_second'] = tmp_dataset[time_col_name].map(lambda x: x.second)
                # Replace column itself with difference from min value, leave as same name
                # to keep index so order of other columns remains the same for other transformations
                tmp_dataset[time_col_name] = tmp_dataset[time_col_name].map(lambda x: x.timestamp() - self._min[idx])
            if not self.return_pandas:
                tmp_dataset = tmp_dataset.values
        return tmp_dataset



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines constants for ml-wrappers."""

from enum import Enum


class ExplanationParams(object):
    """Provide constants for explanation parameters."""

    EXPECTED_VALUES = 'expected_values'
    CLASSES = 'classes'


class ExplainType(object):
    """Provide constants for model and explainer type information, useful for visualization."""

    CLASSIFICATION = 'classification'
    DATA = 'data_type'
    EXPLAIN = 'explain_type'
    EXPLAINER = 'explainer'
    FUNCTION = 'function'
    GLOBAL = 'global'
    HAN = 'han'
    IS_ENG = 'is_engineered'
    IS_RAW = 'is_raw'
    LIME = 'lime'
    LOCAL = 'local'
    METHOD = 'method'
    MIMIC = 'mimic'
    MODEL = 'model_type'
    MODEL_CLASS = 'model_class'
    MODEL_TASK = 'model_task'
    PFI = 'pfi'
    REGRESSION = 'regression'
    SHAP = 'shap'
    SHAP_DEEP = 'shap_deep'
    SHAP_KERNEL = 'shap_kernel'
    SHAP_TREE = 'shap_tree'
    SHAP_LINEAR = 'shap_linear'
    TABULAR = 'tabular'
    SHAP_GPU_KERNEL = 'shap_gpu_kernel'


class ExplainParams(object):
    """Provide constants for interpret community (init, explain_local and explain_global) parameters."""

    BATCH_SIZE = 'batch_size'
    CLASSES = 'classes'
    CLASSIFICATION = 'classification'
    EVAL_DATA = 'eval_data'
    EVAL_Y_PRED = 'eval_y_predicted'
    EVAL_Y_PRED_PROBA = 'eval_y_predicted_proba'
    EXPECTED_VALUES = 'expected_values'
    EXPLAIN_SUBSET = 'explain_subset'
    EXPLANATION_ID = 'explanation_id'
    FEATURES = 'features'
    GLOBAL_IMPORTANCE_NAMES = 'global_importance_names'
    GLOBAL_IMPORTANCE_VALUES = 'global_importance_values'
    GLOBAL_IMPORTANCE_RANK = 'global_importance_rank'
    GLOBAL_NAMES = 'global_names'
    GLOBAL_VALUES = 'global_values'
    GLOBAL_RANK = 'global_rank'
    ID = 'id'
    INCLUDE_LOCAL = 'include_local'
    INIT_DATA = 'init_data'
    IS_ENG = 'is_engineered'
    IS_LOCAL_SPARSE = 'is_local_sparse'
    IS_RAW = 'is_raw'
    LOCAL_EXPLANATION = 'local_explanation'
    LOCAL_IMPORTANCE_VALUES = 'local_importance_values'
    METHOD = 'method'
    MODEL_ID = 'model_id'
    MODEL_TASK = 'model_task'
    MODEL_TYPE = 'model_type'
    NUM_CLASSES = 'num_classes'
    NUM_EXAMPLES = 'num_examples'
    NUM_FEATURES = 'num_features'
    PER_CLASS_NAMES = 'per_class_names'
    PER_CLASS_RANK = 'per_class_rank'
    PER_CLASS_VALUES = 'per_class_values'
    PROBABILITIES = 'probabilities'
    SAMPLING_POLICY = 'sampling_policy'
    SHAP_VALUES_OUTPUT = 'shap_values_output'

    @classmethod
    def get_serializable(cls):
        """Return only the ExplainParams properties that have meaningful data values for serialization.

        :param cls: ExplainParams input class.
        :type cls: ExplainParams
        :return: A set of property names, e.g., 'GLOBAL_IMPORTANCE_VALUES', 'MODEL_TYPE', etc.
        :rtype: set[str]
        """
        return (set(filter(lambda x: not x.startswith('__') and not callable(getattr(cls, x)),
                vars(cls).keys())) - set(['DATA_MAPPER', 'DATA_MAPPER_INTERNAL']))

    @classmethod
    def get_private(cls, explain_param):
        """Return the private version of the ExplainParams property.

        :param cls: ExplainParams input class.
        :type cls: ExplainParams
        :param explain_param: The ExplainParams property to get private version of.
        :type explain_param: str
        :return: The private version of the property.
        :rtype: str
        """
        return '_' + explain_param


class Defaults(object):
    """Provide constants for default values to explain methods."""

    AUTO = 'auto'
    DEFAULT_BATCH_SIZE = 100
    # hdbscan is an unsupervised learning library to find the optimal number of clusters in a dataset
    # See this github repo for more details: https://github.com/scikit-learn-contrib/hdbscan
    HDBSCAN = 'hdbscan'
    MAX_DIM = 50


class Attributes(object):
    """Provide constants for attributes."""

    EXPECTED_VALUE = 'expected_value'


class Dynamic(object):
    """Provide constants for dynamically generated classes."""

    GLOBAL_EXPLANATION = 'DynamicGlobalExplanation'
    LOCAL_EXPLANATION = 'DynamicLocalExplanation'


class Tensorflow(object):
    """Provide TensorFlow and TensorBoard related constants."""

    CPU0 = '/CPU:0'
    TFLOG = 'tflog'


class SKLearn(object):
    """Provide scikit-learn related constants."""

    EXAMPLES = 'examples'
    LABELS = 'labels'
    PREDICTIONS = 'predictions'
    PREDICT = 'predict'
    PREDICT_PROBA = 'predict_proba'


class Spacy(object):
    """Provide spaCy related constants."""

    EN = 'en'
    NER = 'ner'
    TAGGER = 'tagger'


class Scipy(object):
    """Provide scipy related constants."""

    CSR_FORMAT = 'csr'


class ModelTask(str, Enum):
    """Provide model task constants.

    For tabular data, can be 'classification', 'regression', or 'unknown'.
    For text data, can be 'text_classification', 'sentiment_analysis',
    'question_answering', 'entailment', 'summarizations' or 'unknown'.

    By default the model domain is inferred if 'unknown', but this can
    be overridden if you specify 'classification' or 'regression'.
    """

    CLASSIFICATION = 'classification'
    REGRESSION = 'regression'
    TEXT_CLASSIFICATION = 'text_classification'
    MULTILABEL_TEXT_CLASSIFICATION = 'multilabel_text_classification'
    SENTIMENT_ANALYSIS = 'sentiment_analysis'
    QUESTION_ANSWERING = 'question_answering'
    ENTAILMENT = 'entailment'
    SUMMARIZATIONS = 'summarizations'
    IMAGE_CLASSIFICATION = 'image_classification'
    MULTILABEL_IMAGE_CLASSIFICATION = 'multilabel_image_classification'
    OBJECT_DETECTION = 'object_detection'
    UNKNOWN = 'unknown'


class LightGBMParams(object):
    """Provide constants for LightGBM."""

    CATEGORICAL_FEATURE = 'categorical_feature'


class ShapValuesOutput(str, Enum):
    """Provide constants for the SHAP values output from the explainer.

    Can be 'default', 'probability' or 'teacher_probability'. If 'teacher_probability' is specified,
    we use the probabilities from the teacher model.
    """

    DEFAULT = 'default'
    PROBABILITY = 'probability'
    TEACHER_PROBABILITY = 'teacher_probability'


class ExplainableModelType(str, Enum):
    """Provide constants for the explainable model type."""

    TREE_EXPLAINABLE_MODEL_TYPE = 'tree_explainable_model_type'
    LINEAR_EXPLAINABLE_MODEL_TYPE = 'linear_explainable_model_type'


class MimicSerializationConstants(object):
    """Provide internal class that defines fields used for MimicExplainer serialization."""

    FUNCTION = 'function'
    IDENTITY = '_identity'
    INITIALIZATION_EXAMPLES = 'initialization_examples'
    LOGGER = '_logger'
    MODEL = 'model'
    ORIGINAL_EVAL_EXAMPLES = '_original_eval_examples'
    PREDICT_PROBA_FLAG = 'predict_proba_flag'
    TIMESTAMP_FEATURIZER = '_timestamp_featurizer'
    RESET_INDEX = 'reset_index'
    ALLOW_ALL_TRANSFORMATIONS = '_allow_all_transformations'

    enum_properties = ['_shap_values_output']
    nonify_properties = ['_logger', 'model', 'function', 'initialization_examples',
                         '_original_eval_examples', '_timestamp_featurizer']
    save_properties = ['surrogate_model']


class LightGBMSerializationConstants(object):
    """Provide internal class that defines fields used for MimicExplainer serialization."""

    IDENTITY = '_identity'
    LOGGER = '_logger'
    MODEL_STR = 'model_str'
    MULTICLASS = 'multiclass'
    REGRESSION = 'regression'
    TREE_EXPLAINER = '_tree_explainer'
    OBJECTIVE = 'objective'

    enum_properties = ['_shap_values_output']
    nonify_properties = [LOGGER, TREE_EXPLAINER]
    save_properties = ['_lgbm']


class DNNFramework(object):
    """Provide DNN framework constants."""

    TENSORFLOW = 'tensorflow'
    PYTORCH = 'pytorch'


class InterpretData(object):
    """Provide Data and Visualize constants for interpret core."""

    BASE_VALUE = 'Base Value'
    EXPLANATION_TYPE = 'explanation_type'
    EXTRA = 'extra'
    FEATURE_LIST = 'feature_list'
    GLOBAL_FEATURE_IMPORTANCE = 'global_feature_importance'
    INTERCEPT = 'intercept'
    LOCAL_FEATURE_IMPORTANCE = 'local_feature_importance'
    MLI = 'mli'
    NAMES = 'names'
    OVERALL = 'overall'
    PERF = 'perf'
    SCORES = 'scores'
    SPECIFIC = 'specific'
    TYPE = 'type'
    UNIVARIATE = 'univariate'
    VALUE = 'value'
    VALUES = 'values'
    EXPLANATION_CLASS_DIMENSION = 'explanation_class_dimension'
    MULTICLASS = 'multiclass'
    SINGLE = 'single'


class Extension(object):
    """Provide constants for extensions to interpret package."""

    BLACKBOX = 'blackbox'
    GLASSBOX = 'model'
    GLOBAL = 'global'
    GREYBOX = 'specific'
    LOCAL = 'local'


class SHAPDefaults(object):
    """Provide constants for default values to SHAP."""

    INDEPENDENT = 'independent'


class ResetIndex(str, Enum):
    """Provide index column handling constants. Can be 'ignore', 'reset' or 'reset_teacher'.

    By default the index column is ignored, but you can override to reset it and make it a
    feature column that is then featurized to numeric, or reset it and ignore it during
    featurization but set it as the index when calling predict on the original model.
    """

    Ignore = 'ignore'
    Reset = 'reset'
    ResetTeacher = 'reset_teacher'


class Device(Enum):
    """Specifies all possible device types."""

    CPU = 'cpu'
    CUDA = 'cuda'
    AUTO = 'auto'


text_model_tasks = {ModelTask.TEXT_CLASSIFICATION,
                    ModelTask.MULTILABEL_TEXT_CLASSIFICATION,
                    ModelTask.SENTIMENT_ANALYSIS,
                    ModelTask.QUESTION_ANSWERING,
                    ModelTask.ENTAILMENT,
                    ModelTask.SUMMARIZATIONS}


image_model_tasks = {ModelTask.IMAGE_CLASSIFICATION,
                     ModelTask.MULTILABEL_IMAGE_CLASSIFICATION,
                     ModelTask.OBJECT_DETECTION}



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a class for wrapping regression models."""

import pandas as pd
from ml_wrappers.common.constants import ModelTask
from ml_wrappers.model.base_wrapped_model import BaseWrappedModel


class WrappedRegressionModel(BaseWrappedModel):
    """A class for wrapping a regression model."""

    def __init__(self, model, eval_function, examples=None):
        """Initialize the WrappedRegressionModel with the model and evaluation function."""
        super(WrappedRegressionModel, self).__init__(
            model, eval_function, examples, ModelTask.REGRESSION)

    def predict(self, dataset):
        """Predict the output using the wrapped regression model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        preds = self._eval_function(dataset)
        if isinstance(preds, pd.DataFrame):
            preds = preds.values.ravel()

        return preds



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a model wrapper for an endpoint."""

import json
import logging
import os
import ssl
import urllib.request

import numpy as np

module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)

ERR_MSG = 'Could not import mlflow, required if using EndpointWrapperModel'

try:
    from mlflow.pyfunc import PythonModel
    mlflow_installed = True
except ImportError:
    PythonModel = object
    mlflow_installed = False
    module_logger.debug(ERR_MSG)


UTF8 = 'utf8'
IGNORE = 'ignore'
LABEL = 'label'
SCORE = 'score'
# 6 minutes timeout for requests
TIMEOUT = 60 * 6
DEFAULT_BATCH_SIZE = 10


class EndpointWrapperModel(PythonModel):
    """Defines an MLFlow model wrapper for an endpoint."""

    def __init__(self, api_key, url, allow_self_signed_https=False,
                 extra_headers=None, transform_output_dict=False,
                 class_names=None, wrap_input_data_dict=False,
                 batch_size=DEFAULT_BATCH_SIZE,
                 api_key_auto_refresh_callable=None):
        """Initialize the EndpointWrapperModel.

        :param api_key: The API key to use when invoking the endpoint.
        :type api_key: str
        :param url: The URL of the endpoint.
        :type url: str
        :param allow_self_signed_https: Whether to allow certificates.
        :type allow_self_signed_https: bool
        :param extra_headers: Extra headers to send with the request.
        :type extra_headers: dict
        :param transform_output_dict: Whether to transform the output
            dictionary to a numpy array.
        :type transform_output_dict: bool
        :param class_names: The class names.
        :type class_names: list
        :param wrap_input_data_dict: Whether to wrap the input data in a
            dictionary format.
        :type wrap_input_data_dict: bool
        :param batch_size: The batch size to use when invoking the endpoint.
            Calls on full dataset if less than 1.
            This parameter can be used to get around endpoint timeouts.
        :type batch_size: int
        :param api_key_auto_refresh_callable: The method to call to refresh the
            API key.
        :type api_key_auto_refresh_callable: callable
        """
        if not mlflow_installed:
            raise ImportError(ERR_MSG)
        self._api_key = api_key
        if not api_key:
            key_err = "A key should be provided to invoke the endpoint"
            raise ValueError(key_err)
        self._url = url
        self._allow_self_signed_https = allow_self_signed_https
        self._extra_headers = extra_headers
        self._model = self
        self._transform_output_dict = transform_output_dict
        self._class_names = class_names
        self._wrap_input_data_dict = wrap_input_data_dict
        self._batch_size = batch_size
        self._api_key_auto_refresh_callable = api_key_auto_refresh_callable

    @staticmethod
    def from_auto_refresh_callable(api_key_auto_refresh_callable, url, **kwargs):
        """Create an EndpointWrapperModel from an auto refresh callable.

        The callable method should return the latest API key.

        :param api_key_auto_refresh_callable: The method to call to refresh the
            API key.
        :type api_key_auto_refresh_callable: callable
        :param kwargs: The keyword arguments.
        :type kwargs: dict
        :return: The EndpointWrapperModel.
        :rtype: ml_wrappers.model.EndpointWrapperModel
        """
        api_key = api_key_auto_refresh_callable()
        return EndpointWrapperModel(
            api_key,
            url,
            api_key_auto_refresh_callable=api_key_auto_refresh_callable,
            **kwargs)

    def load_context(self, context):
        """Load the context.

        :param context: The context.
        :type context: mlflow.pyfunc.model.PythonModelContext
        """
        # load your artifacts
        pass

    def allow_self_signed_https(self, allowed):
        """Allow self signed HTTPS.

        :param allowed: Whether to allow self signed HTTPS.
        :type allowed: bool
        """
        # bypass the server certificate verification on client side
        verify = os.environ.get('PYTHONHTTPSVERIFY', '')
        ssl_verified = getattr(ssl, '_create_unverified_context', None)
        if allowed and not verify and ssl_verified:
            ssl._create_default_https_context = ssl._create_unverified_context

    def _get_response_with_retry(self, body, timeout, num_retries=3):
        try:
            headers = {'Content-Type': 'application/json',
                       'Authorization': ('Bearer ' + self._api_key)}
            if self._extra_headers:
                headers.update(self._extra_headers)
            req = urllib.request.Request(self._url, body, headers)
            response = urllib.request.urlopen(req, timeout=TIMEOUT)
            json_data = response.read()
        except urllib.error.HTTPError as e:
            print("The request failed with status code: " + str(e.code))
            print("Request body: " + str(body))

            # Print the headers - they include the request ID
            # and the timestamp, which are useful for debugging
            # the failure
            print(e.info())
            # Retry request and refresh API key if refresh method provided
            if num_retries > 0:
                if self._api_key_auto_refresh_callable:
                    self._api_key = self._api_key_auto_refresh_callable()
                return self._get_response_with_retry(
                    body, timeout, num_retries - 1)
            raise ValueError(
                "Request failed with error. " + str(e))
        return json_data

    def _make_request(self, data):
        """Make a request to the endpoint.

        :param data: The data to send to the endpoint.
        :type data: list
        :return: The result.
        :rtype: numpy.ndarray
        """
        if self._wrap_input_data_dict:
            data = {'input_data': data}
        json_input_data = json.dumps(data)
        body = str.encode(json_input_data)
        json_data = self._get_response_with_retry(body, TIMEOUT)
        try:
            result = np.array(json.loads(json_data))
        except Exception as e:
            print("Failed to convert result to numpy array, json_data:")
            print(json_data)
            print(type(json_data))
            raise ValueError(
                "Failed to convert result to numpy array. " + str(e))
        return result

    def _call_webservice(self, data):
        """Common code to call the webservice.

        :param data: The data to send to the webservice.
        :type data: pandas.DataFrame or list
        :return: The result.
        :rtype: numpy.ndarray
        """
        if self._allow_self_signed_https:
            # this line is needed if you use self-signed
            # certificate in your scoring service
            self.allow_self_signed_https(True)

        if not isinstance(data, list):
            if isinstance(data, np.ndarray):
                data = data.tolist()
            else:
                data = data.values.tolist()
        batch_size = self._batch_size
        lt_batch_size = isinstance(data, list) and len(data) <= batch_size
        request_all = batch_size < 1 or lt_batch_size
        if request_all:
            return self._make_request(data)
        else:
            result = []
            for i in range(0, len(data), batch_size):
                batch_result = self._make_request(data[i:i + batch_size])
                result.extend(batch_result.tolist())
            return np.array(result)

    def predict(self, context, model_input=None):
        """Predict using the model.

        :param context: The context for MLFlow model or the input data.
        :type context: mlflow.pyfunc.model.PythonModelContext or
            pandas.DataFrame
        :param model_input: The input to the model.
        :type model_input: pandas.DataFrame
        :return: The predictions.
        :rtype: numpy.ndarray
        """
        # This is to conform to the scikit-learn API format
        # which MLFlow does not follow
        if model_input is None:
            model_input = context
        result = self._call_webservice(model_input)
        if self._transform_output_dict:
            array_result = []
            for row in result:
                label = row[LABEL]
                prediction = self._class_names.index(label)
                array_result.append(prediction)
            result = np.array(array_result)
        return result

    def predict_proba(self, data):
        """Predict using the model.

        :param data: The input to the model.
        :type data: pandas.DataFrame
        :return: The predictions.
        :rtype: numpy.ndarray
        """
        result = self._call_webservice(data)
        if self._transform_output_dict:
            array_result = []
            for row in result:
                label = row[LABEL]
                score = row[SCORE]
                if label == self._class_names[0]:
                    array_result.append([score, 1 - score])
                else:
                    array_result.append([1 - score, score])
            result = np.array(array_result)
        return result

    def __call__(self, data):
        """Call the model.

        :param data: The input to the model.
        :type data: pandas.DataFrame
        :return: The predictions.
        :rtype: numpy.ndarray
        """
        return self.predict_proba(data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines helpful utilities for the DatasetWrapper."""

import logging

import numpy as np
from scipy.sparse import csr_matrix, issparse
from scipy.sparse import vstack as sparse_vstack
from sklearn.utils import shuffle
from sklearn.utils.sparsefuncs import csc_median_axis_0

from ..common.gpu_kmeans import kmeans
from ..common.warnings_suppressor import shap_warnings_suppressor

with shap_warnings_suppressor():
    try:
        import shap
        shap_installed = True
    except BaseException:  # noqa: B036
        shap_installed = False

module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


def _generate_augmented_data(x, max_num_of_augmentations=np.inf):
    """Augment x by appending x with itself shuffled columnwise many times.

    :param x: data that has to be augmented, array or sparse matrix of 2 dimensions
    :type x: numpy.ndarray or scipy.sparse.csr_matrix
    :param max_augment_data_size: number of times we stack permuted x to augment.
    :type max_augment_data_size: int
    :return: augmented data with roughly number of rows that are equal to number of columns
    :rtype: numpy.ndarray or scipy.sparse.csr_matrix
    """
    x_augmented = x
    vstack = sparse_vstack if issparse(x) else np.vstack
    for i in range(min(x.shape[1] // x.shape[0] - 1, max_num_of_augmentations)):
        x_permuted = shuffle(x.T, random_state=i).T
        x_augmented = vstack([x_augmented, x_permuted])

    return x_augmented


def _summarize_data(X, k=10, use_gpu=False, to_round_values=True):
    """Summarize a dataset.

    For dense dataset, use k mean samples weighted by the number of data points they
    each represent.
    For sparse dataset, use a sparse row for the background with calculated
    median for dense columns.

    :param X: Matrix of data samples to summarize (# samples x # features).
    :type X: numpy.ndarray or pandas.DataFrame or scipy.sparse.csr_matrix
    :param k: Number of cluster centroids to use for approximation.
    :type k: int
    :param to_round_values: When using kmeans, for each element of every cluster centroid to match the nearest value
        from X in the corresponding dimension. This ensures discrete features
        always get a valid value.  Ignored for sparse data sample.
    :type to_round_values: bool
    :return: summarized numpy array or csr_matrix object.
    :rtype: numpy.ndarray or scipy.sparse.csr_matrix or DenseData
    """
    is_sparse = issparse(X)
    if not str(type(X)).endswith(".DenseData'>"):
        if is_sparse:
            module_logger.debug('Creating sparse data summary as csr matrix')
            # calculate median of sparse background data
            median_dense = csc_median_axis_0(X.tocsc())
            return csr_matrix(median_dense)
        elif len(X) > 10 * k:
            module_logger.debug('Create dense data summary with k-means')
            # use kmeans to summarize the examples for initialization
            # if there are more than 10 x k of them
            if use_gpu:
                return kmeans(X, k, to_round_values)
            else:
                if not shap_installed:
                    raise RuntimeError('shap is required to compute dataset summary in DatasetWrapper')
                return shap.kmeans(X, k, to_round_values)
    return X


def _convert_batch_dataset_to_numpy(batch_dataset):
    """Convert a TensorFlow batch dataset to a numpy array.

    :param batch_dataset: batch dataset to convert
    :type batch_dataset: BatchDataset
    :return: data, feature names and batch size
    :rtype: numpy.ndarray, list, int
    """
    batches = []
    set_keys = False
    features = []
    batch_size = 0
    for data, _ in batch_dataset:
        columns = []
        for column in data.values():
            columns.append(np.array(column))
        if not set_keys:
            for key in data.keys():
                features.append(key)
            batch_size = columns[0].shape[0]
            set_keys = True
        batches.append(np.stack(columns, axis=1))
    converted_data = np.vstack(batches)
    return converted_data, features, batch_size



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines classes to wrap the training/test data and the corresponding predictions from the model."""

from typing import Optional

import numpy as np
import pandas as pd

TARGET = 'target'


class PredictionsModelWrapper:
    """Model wrapper to wrap the samples used to train the models
    and the predictions of the model. This wrapper is useful when
    it is not possible to load the model.
    """
    def __init__(self, test_data: pd.DataFrame, y_pred: np.ndarray,
                 should_construct_pandas_query: Optional[bool] = True):
        """Creates a PredictionsModelWrapper object.

        :param test_data: The data that was used to train the model.
        :type test_data: pd.DataFrame
        :param y_pred: Predictions of the model.
        :type y_pred: np.ndarray
        :param should_construct_pandas_query: Whether to use pandas query()
            function to filter the data. If set to False, the data will
            be filtered using iloc(). Defaults to True.
        :type should_construct_pandas_query: Optional[bool]
        """
        if not isinstance(test_data, pd.DataFrame):
            raise DataValidationException(
                "Expecting a pandas dataframe for test_data")
        if not isinstance(y_pred, np.ndarray):
            raise DataValidationException(
                "Expecting a numpy array for y_pred")
        if len(test_data) != len(y_pred):
            raise DataValidationException(
                "The number of instances in test data "
                "do not match with number of predictions")
        self._feature_names = list(test_data.columns)
        self._combined_data = test_data.copy()
        self._combined_data[TARGET] = y_pred
        self._should_construct_pandas_query = should_construct_pandas_query

    def _get_filtered_data(
            self, query_test_data_row: pd.DataFrame) -> pd.DataFrame:
        """Return the filtered data based on the query data.

        :param query_test_data_row: The query data instance.
        :type query_test_data_row: pd.DataFrame
        :return: The filtered dataframe based on the values in query data.
        :rtype: pd.DataFrame
        """
        if not self._should_construct_pandas_query:
            data_copy = self._combined_data
            for column_name, column_data in query_test_data_row.squeeze().items():
                if pd.isnull(column_data):
                    data_copy = data_copy[data_copy[column_name].isna()]
                else:
                    data_copy = data_copy[data_copy[column_name] == column_data]

                if len(data_copy) == 1 or len(data_copy) == 0:
                    # Stop the search if we have found a possible exact match or no match
                    break

            if len(data_copy) == 1:
                if query_test_data_row.equals(data_copy.filter(self._feature_names)):
                    filtered_df = data_copy
                else:
                    filtered_df = pd.DataFrame(columns=self._combined_data.columns)
            else:
                filtered_df = data_copy
        else:
            queries = []
            for column_name, column_data in query_test_data_row.squeeze().items():
                if pd.isnull(column_data):
                    queries.append("`{}`.isna()".format(column_name))
                elif isinstance(column_data, str):
                    queries.append("`{}` == '{}'".format(
                        column_name, column_data))
                else:
                    queries.append("`{}` == {}".format(
                        column_name, column_data))

            queries_str = '(' + ') & ('.join(queries) + ')'
            filtered_df = self._combined_data.query(queries_str)

        if len(filtered_df) == 0:
            raise EmptyDataException(
                "The query data was not found in the combined dataset")

        return filtered_df

    def predict(self, query_test_data: pd.DataFrame) -> np.ndarray:
        """Return the predictions based on the query data.

        :param query_test_data: The data for which the predictions need to
            be returned.
        :type query_test_data: pd.DataFrame
        :return: Predictions of the model.
        :rtype: np.ndarray
        """
        if not isinstance(query_test_data, pd.DataFrame):
            raise DataValidationException(
                "Expecting a pandas dataframe for query_test_data")
        prediction_output = []
        for index in range(0, len(query_test_data)):
            filtered_df = self._get_filtered_data(
                query_test_data[index:index + 1])
            prediction_output.append(filtered_df[TARGET].values[0])

        return np.array(prediction_output)

    def __setstate__(self, state):
        """Set the state of the class object so that the wrapped
        model is serializable via pickle.

        :param state: A dictionary of deserialized state.
        :type state: Dict
        """
        self._combined_data = state["_combined_data"]
        self._should_construct_pandas_query = state["_should_construct_pandas_query"]
        self._feature_names = state["_feature_names"]

    def __getstate__(self):
        """Return the state so that the wrapped model is
        serializable via pickle.

        :return: The state to be pickled.
        :rtype: Dict
        """
        state = {}
        state["_combined_data"] = self._combined_data
        state["_should_construct_pandas_query"] = self._should_construct_pandas_query
        state["_feature_names"] = self._feature_names
        return state


class PredictionsModelWrapperRegression(PredictionsModelWrapper):
    """Model wrapper to wrap the samples used to train the models
    and the predictions of the model for regression tasks.
    """
    def __init__(self, test_data: pd.DataFrame, y_pred: np.ndarray,
                 should_construct_pandas_query: Optional[bool] = True):
        """Creates a PredictionsModelWrapperRegression object.

        :param test_data: The data that was used to train the model.
        :type test_data: pd.DataFrame
        :param y_pred: Predictions of the model.
        :type y_pred: np.ndarray
        :param should_construct_pandas_query: Whether to use pandas query()
            function to filter the data. If set to False, the data will
            be filtered using iloc(). Defaults to True.
        :type should_construct_pandas_query: Optional[bool]
        """
        super(PredictionsModelWrapperRegression, self).__init__(
            test_data, y_pred, should_construct_pandas_query)


class PredictionsModelWrapperClassification(PredictionsModelWrapper):
    """Model wrapper to wrap the samples used to train the models
    and the predictions of the model for classification tasks.
    """
    def __init__(self, test_data: pd.DataFrame, y_pred: np.ndarray,
                 y_pred_proba: Optional[np.ndarray] = None,
                 should_construct_pandas_query: Optional[bool] = True):
        """Creates a PredictionsModelWrapperClassification object.

        :param test_data: The data that was used to train the model.
        :type test_data: pd.DataFrame
        :param y_pred: Predictions of the model.
        :type y_pred: np.ndarray
        :param y_pred_proba: Prediction probabilities of the model.
        :type y_pred_proba: np.ndarray
        :param should_construct_pandas_query: Whether to use pandas query()
            function to filter the data. If set to False, the data will
            be filtered using iloc(). Defaults to True.
        :type should_construct_pandas_query: Optional[bool]
        """
        super(PredictionsModelWrapperClassification, self).__init__(
            test_data, y_pred, should_construct_pandas_query)
        self._num_classes = None
        if y_pred_proba is not None:
            if not isinstance(y_pred_proba, np.ndarray):
                raise DataValidationException(
                    "Expecting a numpy array for y_pred_proba")

            if len(test_data) != len(y_pred_proba):
                raise DataValidationException(
                    "The number of instances in test data "
                    "do not match with number of prediction probabilities")

            for i in range(0, len(y_pred_proba[0])):
                self._combined_data[
                    TARGET + '_' + str(i)] = y_pred_proba[:, i]
            self._num_classes = len(y_pred_proba[0])

    def predict_proba(self, query_test_data: pd.DataFrame) -> np.ndarray:
        """Return the prediction probabilities based on the query data.

        :param query_test_data: The data for which the prediction
            probabilities need to be returned.
        :type query_test_data: pd.DataFrame
        :return: Prediction probabilities of the model.
        :rtype: np.ndarray
        """
        if not isinstance(query_test_data, pd.DataFrame):
            raise DataValidationException(
                "Expecting a pandas dataframe for query_test_data")
        if self._num_classes is None:
            raise DataValidationException(
                "Model wrapper configured without prediction probabilities"
            )
        prediction_output = []
        for index in range(0, len(query_test_data)):
            filtered_df = self._get_filtered_data(
                query_test_data[index:index + 1])
            classes_output = []
            for i in range(self._num_classes):
                classes_output.append(
                    filtered_df[TARGET + '_' + str(i)].values[0])
            prediction_output.append(classes_output)

        return np.array(prediction_output)

    def __setstate__(self, state):
        """Set the state of the class object so that the wrapped
        model is serializable via pickle.

        :param state: A dictionary of deserialized state.
        :type state: Dict
        """
        super().__setstate__(state)
        self._num_classes = state["_num_classes"]

    def __getstate__(self):
        """Return the state so that the wrapped model is
        serializable via pickle.

        :return: The state to be pickled.
        :rtype: Dict
        """
        state = super().__getstate__()
        state["_num_classes"] = self._num_classes
        return state


class EmptyDataException(Exception):
    """An exception indicating that some operation produced empty data.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = 'Empty data exception'


class DataValidationException(Exception):
    """An exception indicating that some user supplied data is not valid.

    :param exception_message: A message describing the error.
    :type exception_message: str
    """
    _error_code = 'Invalid data'



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines model wrappers and utilities for tensorflow models."""

import numpy as np
import pandas as pd

PREDICT_CLASSES = 'predict_classes'


def is_sequential(model):
    """Returns True if the model is a sequential model.

    Note the model class name can be
    keras.src.engine.sequential.Sequential,
    keras.engine.sequential.Sequential or
    tensorflow.python.keras.engine.sequential.Sequential
    depending on the tensorflow version.
    In the latest 2.13 version, the namespace changed from
    keras.engine to keras.src.engine.
    The check should include all of these cases.

    :param model: The model to check.
    :type model: tf.keras.Model
    :return: True if the model is a sequential model.
    :rtype: bool
    """
    model_type = str(type(model))
    old_sequential_ns = "keras.engine.sequential.Sequential'>"
    # the sequential namespace changed in tensorflow 2.13
    new_sequential_ns1 = "keras.src.engine.sequential.Sequential'>"
    # it changed again in tensorflow 2.17
    new_sequential_ns2 = "keras.src.models.sequential.Sequential'>"
    return any([model_type.endswith(old_sequential_ns),
                model_type.endswith(new_sequential_ns1),
                model_type.endswith(new_sequential_ns2)])


class WrappedTensorflowModel(object):
    """A class for wrapping a TensorFlow model.

    Note at time of initialization, since we don't have
    access to the dataset, we can't infer if this is for
    classification or regression case.  Hence, we add
    the predict_classes method for classification, and keep
    predict for either outputting values in regression or
    probabilities in classification.
    """

    def __init__(self, model):
        """Initialize the WrappedTensorflowModel with the model.

        :param model: The model to wrap.
        :type model: tf.keras.Model
        """
        self._model = model

    def predict(self, dataset):
        """Predict the output using the wrapped TensorFlow model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The prediction results.
        :rtype: numpy.ndarray
        """
        # Convert the data to numpy
        if isinstance(dataset, pd.DataFrame):
            dataset = dataset.values
        return self._model.predict(dataset)

    def predict_classes(self, dataset):
        """Predict the class using the wrapped TensorFlow model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted classes.
        :rtype: numpy.ndarray
        """
        # Note predict_classes was removed for models after
        # tensorflow version 2.6
        if hasattr(self._model, PREDICT_CLASSES):
            return self._model.predict_classes(dataset)
        probabilities = self.predict_proba(dataset)
        return np.argmax(probabilities, axis=1)

    def predict_proba(self, dataset):
        """Predict the output probability using the wrapped TensorFlow model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted probabilities.
        :rtype: numpy.ndarray
        """
        return self.predict(dataset)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Setup file for ml-wrappers package."""
import os
import shutil

from setuptools import find_packages, setup

with open('ml_wrappers/version.py') as f:
    code = compile(f.read(), f.name, 'exec')
    exec(code)

README_FILE = 'README.md'
LICENSE_FILE = 'LICENSE.txt'

# Note: used when generating the wheel but not on pip install of the package
if os.path.exists('../LICENSE'):
    shutil.copyfile('../LICENSE', LICENSE_FILE)


CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: MacOS',
    'Operating System :: POSIX :: Linux'
]

DEPENDENCIES = [
    'numpy<2.0.0',
    'packaging',
    'pandas<2.0.0',
    'scipy',
    'scikit-learn'
]

with open(README_FILE, 'r', encoding='utf-8') as f:
    README = f.read()

setup(
    name=name,  # noqa: F821
    version=version,  # noqa: F821
    description='Machine Learning Wrappers SDK for Python',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Microsoft Corp',
    author_email='ilmat@microsoft.com',
    license='MIT License',
    url='https://github.com/microsoft/ml-wrappers',
    classifiers=CLASSIFIERS,
    packages=find_packages(exclude=["*.tests"]),
    install_requires=DEPENDENCIES,
    zip_safe=False
)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a class for wrapping classifiers without predict_proba."""

import numpy as np


class WrappedClassificationWithoutProbaModel(object):
    """A class for wrapping a classifier without a predict_proba method.

    Note: the classifier may not output numeric values for its predictions.
    We generate a trival boolean version of predict_proba
    """

    def __init__(self, model):
        """Initialize the WrappedClassificationWithoutProbaModel with the model."""
        self._model = model
        # Create a map from classes to index
        self._classes_to_index = {}
        for index, i in enumerate(self._model.classes_):
            self._classes_to_index[i] = index
        self._num_classes = len(self._model.classes_)

    def predict(self, dataset):
        """Predict the output using the wrapped regression model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        return self._model.predict(dataset)

    def predict_proba(self, dataset):
        """Predict the output probability using the wrapped model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        """
        predictions = self.predict(dataset)
        # Generate trivial boolean array for predictions
        probabilities = np.zeros((predictions.shape[0], self._num_classes))
        for row_idx, pred_class in enumerate(predictions):
            class_index = self._classes_to_index[pred_class]
            probabilities[row_idx, class_index] = 1
        return probabilities



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines helpful model wrapper and utils for implicitly rewrapping the model
    to conform to explainer contracts."""

import logging
import warnings
from typing import Union

import numpy as np
from ml_wrappers.model.wrapped_classification_model import \
    WrappedClassificationModel
from ml_wrappers.model.wrapped_classification_without_proba_model import \
    WrappedClassificationWithoutProbaModel
from ml_wrappers.model.wrapped_regression_model import WrappedRegressionModel
from sklearn.linear_model import SGDClassifier

from ..common.constants import (Device, ModelTask, SKLearn, image_model_tasks,
                                text_model_tasks)
from ..dataset.dataset_wrapper import DatasetWrapper
from .evaluator import _eval_function, _eval_model
from .fastai_wrapper import WrappedFastAITabularModel, _is_fastai_tabular_model
from .image_model_wrapper import _wrap_image_model
from .pytorch_wrapper import WrappedPytorchModel
from .tensorflow_wrapper import WrappedTensorflowModel, is_sequential
from .text_model_wrapper import _is_transformers_pipeline, _wrap_text_model

with warnings.catch_warnings():
    warnings.filterwarnings(
        'ignore', 'Starting from version 2.2.1', UserWarning)


module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


try:
    import torch.nn as nn
except ImportError:
    module_logger.debug(
        'Could not import torch, required if using a PyTorch model')


def wrap_model(model, examples, model_task: str = ModelTask.UNKNOWN,
               num_classes: int = None, classes: Union[list, np.array] = None,
               device=Device.AUTO.value):
    """If needed, wraps the model in a common API based on model task and
        prediction function contract.

    :param model: The model to evaluate on the examples.
    :type model: model with a predict or predict_proba function.
    :param examples: The model evaluation examples.
        Note the examples will be wrapped in a DatasetWrapper, if not
        wrapped when input.
    :type examples: ml_wrappers.DatasetWrapper or numpy.ndarray
        or pandas.DataFrame or panads.Series or scipy.sparse.csr_matrix
        or shap.DenseData or torch.Tensor
    :param model_task: Optional parameter to specify whether the model
                       is a classification or regression model.
                       In most cases, the type of the model can be inferred
                       based on the shape of the output, where a classifier
                       has a predict_proba method and outputs a 2 dimensional
                       array, while a regressor has a predict method and
                       outputs a 1 dimensional array.
    :param classes: optional parameter specifying a list of class names
        the dataset
    :type classes: list or np.ndarray
    :param num_classes: optional parameter specifying the number of classes in
        the dataset
    :type num_classes: int
    :type model_task: str
    :param device: optional parameter specifying the device to move the model
        to. If not specified, then cpu is the default
    :type device: str, for instance: 'cpu', 'cuda'
    :return: The wrapper model.
    :rtype: model
    """
    if model_task == ModelTask.UNKNOWN and _is_transformers_pipeline(model):
        # TODO: can we also dynamically figure out the task if it was
        # originally unknown for text scenarios?
        raise ValueError("ModelTask must be specified for text-based models")
    if model_task in text_model_tasks:
        return _wrap_text_model(model, examples, model_task, False)[0]
    if model_task in image_model_tasks:
        return _wrap_image_model(model, examples, model_task,
                                 False, num_classes, classes,
                                 device)[0]
    return _wrap_model(model, examples, model_task, False)[0]


def _wrap_model(model, examples, model_task, is_function):
    """If needed, wraps the model or function in a common API based on model
        task and prediction function contract.

    :param model: The model or function to evaluate on the examples.
    :type model: function or model with a predict or predict_proba function
    :param examples: The model evaluation examples.
        Note the examples will be wrapped in a DatasetWrapper, if not
        wrapped when input.
    :type examples: ml_wrappers.DatasetWrapper or numpy.ndarray
        or pandas.DataFrame or panads.Series or scipy.sparse.csr_matrix
        or shap.DenseData or torch.Tensor
    :param model_task: Optional parameter to specify whether the model
                       is a classification or regression model.
                       In most cases, the type of the model can be inferred
                       based on the shape of the output, where a classifier
                       has a predict_proba method and outputs a 2 dimensional
                       array, while a regressor has a predict method and
                       outputs a 1 dimensional array.
    :type model_task: str
    :return: The function chosen from given model and chosen domain, or
             model wrapping the function and chosen domain.
    :rtype: (function, str) or (model, str)
    """
    if not isinstance(examples, DatasetWrapper):
        examples = DatasetWrapper(examples)
    if is_function:
        return _eval_function(model, examples, model_task)
    else:
        try:
            if isinstance(model, nn.Module):
                # Wrap the model in an extra layer that
                # converts the numpy array

                # to pytorch Variable and adds predict and
                # predict_proba functions
                model = WrappedPytorchModel(model)
        except (NameError, AttributeError):
            module_logger.debug(
                'Could not import torch, required if using a pytorch model')
        if _is_fastai_tabular_model(model):
            model = WrappedFastAITabularModel(model)
        if is_sequential(model):
            model = WrappedTensorflowModel(model)
        if _classifier_without_proba(model):
            model = WrappedClassificationWithoutProbaModel(model)
        eval_function, eval_ml_domain = _eval_model(
            model, examples, model_task)
        if eval_ml_domain == ModelTask.CLASSIFICATION:
            return WrappedClassificationModel(model, eval_function, examples), \
                eval_ml_domain
        else:
            return WrappedRegressionModel(model, eval_function, examples), \
                eval_ml_domain


def _classifier_without_proba(model):
    """Returns True if the given model is a classifier without predict_proba,
        eg SGDClassifier.

    :param model: The model to evaluate on the examples.
    :type model: model with a predict or predict_proba function
    :return: True if the given model is a classifier without predict_proba.
    :rtype: bool
    """
    return isinstance(model, SGDClassifier) and not \
        hasattr(model, SKLearn.PREDICT_PROBA)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines helper utilities for resolving prediction function shape inconsistencies."""

import logging

import numpy as np

from ..common.constants import ModelTask

module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


try:
    import torch
    pytorch_installed = True
except ImportError:
    pytorch_installed = False
    module_logger.debug('Could not import torch, required if using a PyTorch model')


def _convert_to_two_cols(function, examples):
    """In classification case, convert the function's output to two columns if it outputs one column.

    :param function: The prediction function to evaluate on the examples.
    :type function: function
    :param examples: The model evaluation examples.
    :type examples: numpy.ndarray or list
    :return: The function chosen from given model and classification domain.
    :rtype: (function, str)
    """
    # Add wrapper function to convert output to 2D array, check values to decide on whether
    # to throw, or create two columns [1-p, p], or leave just one in multiclass one-class edge-case
    result = function(examples)
    single_row_result = function(examples[0:1])
    is_single_row_1d_result = len(single_row_result.shape) == 1
    # If the function gives a 2D array of one column, we will need to reshape it prior to concat
    is_2d_result = len(result.shape) == 2
    is_multi_row_multi_cols = result.shape[1] > 1
    reshape_1_row_1d = is_single_row_1d_result and is_2d_result and is_multi_row_multi_cols
    wrapper = _FunctionWrapper(function)
    if reshape_1_row_1d:
        # special weird case for multiclass KerasClassifier where single
        # row result is 1D but multi row result is 2D, even when passing
        # a single row as either a 1D or 2D array
        return (wrapper._function_2D_single_row_wrapper_2D_result, ModelTask.CLASSIFICATION)
    convert_to_two_cols = False
    for value in result:
        if value < 0 or value > 1:
            error = "Probability values outside of valid range: "
            error += str(value)
            raise Exception(error)
        if value < 1:
            convert_to_two_cols = True
    if convert_to_two_cols:
        # Create two cols, [1-p, p], from evaluation result
        if is_2d_result:
            return (wrapper._function_2D_two_cols_wrapper_2D_result, ModelTask.CLASSIFICATION)
        else:
            return (wrapper._function_2D_two_cols_wrapper_1D_result, ModelTask.CLASSIFICATION)
    else:
        if is_2d_result:
            return (function, ModelTask.CLASSIFICATION)
        else:
            return (wrapper._function_2D_one_col_wrapper, ModelTask.CLASSIFICATION)


class _FunctionWrapper(object):
    """Wraps a function to reshape the input and output data.

    :param function: The prediction function to evaluate on the examples.
    :type function: function
    """

    def __init__(self, function, base_dims=1, is_pytorch_image_model=False):
        """Wraps a function to reshape the input and output data.

        :param function: The prediction function to evaluate on the examples.
        :type function: function
        :param base_dims: The base dimensions of the input data.
        :type base_dims: int
        """
        self._function = function
        self._base_dims = base_dims
        self._is_pytorch_image_model = is_pytorch_image_model

    def _function_input_expand_wrapper(self, dataset):
        """Wraps a function that expands the dims of input dataset.

        Note unlike other functions, this runs on the input dataset type,
        and not the converted type.  Note in practice only numpy and tensor
        arrays have been seen to have this issue for models, this function
        does not apply to pandas dataframes.

        :param dataset: The model evaluation examples.
        :type dataset: numpy.ndarray or torch.Tensor
        :return: A wrapped function.
        :rtype: function
        """
        if len(dataset.shape) == self._base_dims:
            is_tensor = False
            if pytorch_installed and isinstance(dataset, torch.Tensor):
                is_tensor = True
            # pytorch requires extra dimension for grayscale images
            # this only needs to be done if image is not yet in tensor format
            is_2d = len(dataset.shape) == 2
            if not is_tensor and self._is_pytorch_image_model and is_2d:
                dataset = np.expand_dims(dataset, axis=2)
            dataset = np.expand_dims(dataset, axis=0)
            if is_tensor:
                dataset = torch.tensor(dataset)
        return self._function(dataset)

    def _function_flatten(self, dataset):
        """Wraps a function that flattens the input dataset from 2D to 1D.

        :param dataset: The model evaluation examples.
        :type dataset: numpy.ndarray
        :return: A wrapped function.
        :rtype: function
        """
        return self._function(dataset).flatten()

    def _function_2D_two_cols_wrapper_2D_result(self, dataset):
        """Wraps a function that creates two columns, [1-p, p], from 2D array of one column evaluation result.

        :param dataset: The model evaluation examples.
        :type dataset: numpy.ndarray
        :return: A wrapped function.
        :rtype: function
        """
        result = self._function(dataset)[:, 0]
        return np.stack([1 - result, result], axis=-1)

    def _function_2D_two_cols_wrapper_1D_result(self, dataset):
        """Wraps a function that creates two columns, [1-p, p], from evaluation result that is a 1D array.

        :param dataset: The model evaluation examples.
        :type dataset: numpy.ndarray
        :return: A wrapped function.
        :rtype: function
        """
        result = self._function(dataset)
        return np.stack([1 - result, result], axis=-1)

    def _function_2D_one_col_wrapper(self, dataset):
        """Wraps a function that creates one column in rare edge case scenario for multiclass one-class result.

        :param dataset: The model evaluation examples.
        :type dataset: numpy.ndarray
        :return: A wrapped function.
        :rtype: function
        """
        result = self._function(dataset)
        return result.reshape(result.shape[0], 1)

    def _function_2D_single_row_wrapper_2D_result(self, dataset):
        """Wraps a single row result in rare multiclass case.

        In this rare multiclass case the single row result
        is 1D but multi row result is 2D.  The single row,
        even when formatted as a 1D or 2D array, gives similar
        consistent 1D array result, but as soon more than 2
        rows are given the result is a 2D array.

        :param dataset: The model evaluation examples.
        :type dataset: numpy.ndarray
        :return: A wrapped function.
        :rtype: function
        """
        result = self._function(dataset)
        if dataset.shape[0] == 1:
            return result.reshape(1, result.shape[0])
        return result


class _MultiVsSingleInstanceFunctionResolver(object):
    """Wraps function output to be same in single and multi-instance cases."""

    def __init__(self, function):
        """Wraps function output to be same in single and multi-instance cases

        :param function: The prediction function to evaluate on the examples.
        :type function: function
        """
        self._function = function

    def _add_single_predict_dimension(self, dataset):
        """Wraps function output for single instance case to add dimension.

        Ensures the single instance case returns same dimensionality
        as multi-instance prediction.

        :param dataset: The model evaluation examples.
        :type dataset: numpy.ndarray
        :return: A wrapped function.
        :rtype: function
        """
        result = self._function(dataset)
        if len(dataset.shape) == 1:
            return np.expand_dims(result, axis=0)
        else:
            return result



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines wrappers for vision-based models."""

import logging
from typing import Any, Dict, Tuple, Union

import numpy as np
import pandas as pd
from ml_wrappers.common.constants import Device, ModelTask
from ml_wrappers.dataset.dataset_wrapper import DatasetWrapper
from ml_wrappers.model.evaluator import _eval_model
from ml_wrappers.model.model_utils import (_is_callable_pipeline,
                                           _is_transformers_pipeline)
from ml_wrappers.model.pytorch_wrapper import WrappedPytorchModel
from ml_wrappers.model.wrapped_classification_model import \
    WrappedClassificationModel

module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)

try:
    import torch
    import torch.nn as nn
    from torch import Tensor
except ImportError:
    Tensor = Any
    module_logger.debug('Could not import torch, required if using a '
                        'PyTorch model')

try:
    from vision_explanation_methods.explanations import common as od_common
    from vision_explanation_methods.explanations.common import \
        GeneralObjectDetectionModelWrapper
except ImportError:
    GeneralObjectDetectionModelWrapper = object
    module_logger.debug('Could not import vision_explanation_methods, '
                        'required if using PytorchDRiseWrapper')

try:
    import torchvision
    from torchvision import transforms as T
except ImportError:
    module_logger.debug('Could not import torchvision, required if '
                        'using PytorchDRiseWrapper')

try:
    from mlflow.pyfunc import PyFuncModel
except ImportError:
    PyFuncModel = Any
    module_logger.debug('Could not import mlflow, required if using an '
                        'mlflow model')

FASTAI_MODEL_SUFFIX = "fastai.learner.Learner'>"
BOXES = 'boxes'
LABELS = 'labels'
SCORES = 'scores'
COLON = ":"


def _is_fastai_model(model):
    """Check if the model is a fastai model.

    :param model: The model to check.
    :type model: object
    :return: True if the model is a fastai model, False otherwise.
    :rtype: bool
    """
    return str(type(model)).endswith(FASTAI_MODEL_SUFFIX)


def _filter_score(orig_prediction: dict, score_threshold: float = 0.5):
    """Filter out predictions with confidence scores < score_threshold.

    :param orig_prediction: Original model prediction
    :type orig_prediction: dict
    :param score_threshold: Score threshold to filter by
    :type score_threshold: float
    :return: Model predictions filtered out by score_threshold
    :rtype: dict
    """
    keep = orig_prediction[SCORES] > score_threshold

    filter_prediction = orig_prediction
    filter_prediction[BOXES] = filter_prediction[BOXES][keep]
    filter_prediction[SCORES] = filter_prediction[SCORES][keep]
    filter_prediction[LABELS] = filter_prediction[LABELS][keep]
    return filter_prediction


def _apply_nms(orig_prediction: dict, iou_threshold: float = 0.5):
    """Perform nms on the predictions based on their IoU.

    :param orig_prediction: Original model prediction
    :type orig_prediction: dict
    :param iou_threshold: iou_threshold for nms
    :type iou_threshold: float
    :return: Model prediction after nms is applied
    :rtype: dict
    """
    nms_prediction = orig_prediction
    boxes = orig_prediction[BOXES]
    scores = orig_prediction[SCORES]
    if boxes.numel() != 0:
        keep = torchvision.ops.nms(boxes,
                                   scores,
                                   iou_threshold)

        nms_prediction[BOXES] = boxes[keep]
        nms_prediction[SCORES] = scores[keep]
        nms_prediction[LABELS] = nms_prediction[LABELS][keep]
    return nms_prediction


def _process_automl_detections_to_raw_detections(
        image_detections,
        label_dict: Dict[str, int],
        image_size: Tuple[int, int]) -> Dict[str, Tensor]:
    """Process AutoML mlflow object detections from a single image.

    The length of image_detections list will be equal to the number
    of objects the AutoML MLflow model detected in a single image.
    Below is an example of image_detections when the model detected
    two objects:

    [{'box': {'topX': 0.645,
              'topY': 0.373,
              'bottomX': 0.8276,
              'bottomY': 0.6102
                },
      'label': 'can',
      'score': 0.945},
     {'box': {'topX': 0.324,
              'topY': 0.5643,
              'bottomX': 0.6511,
              'bottomY': 0.865
                },
      'label': 'milk_bottle',
      'score': 0.93}
    ]
    The bbox coordinates need to be scaled by the image dimensions.

    :param image_detections: AutoML mlflow detections
    :type image_detections: list
    :param label_dict: Label dictionary mapping class names to an index
    :type label_dict: dict
    :param image_size: Image size
    :type image_size: tuple(int, int)
    :return: Raw detections
    :rtype: dict
    """

    x, y = image_size

    boxes = []
    scores = []
    labels = []
    for detection in image_detections:
        # predicted class label
        label = detection['label']
        # confidence score from model
        score = detection['score']
        # predicted bbox coordinates
        box = detection["box"]

        ymin, xmin, ymax, xmax = (
            box["topY"], box["topX"], box["bottomY"], box["bottomX"])

        # the automl mlflow model generates normalized bbox coordinates
        # they need to be scaled by the size of the image
        x_min_scaled, y_min_scaled = x * xmin, y * ymin
        x_max_scaled, y_max_scaled = x * xmax, y * ymax

        scores.append(score)
        labels.append(label)
        boxes.append([x_min_scaled, y_min_scaled, x_max_scaled, y_max_scaled])
    try:
        labels = [int(x) for x in labels]
    except Exception:
        labels = [label_dict[x] for x in labels]

    return {
        BOXES: Tensor(boxes),
        LABELS: Tensor(labels),
        SCORES: Tensor(scores)}


def expand_class_scores(
        scores: Tensor,
        labels: Tensor,
        number_of_classes: int,
) -> Tensor:
    """Extrapolate a full set of class scores.

    Many object detection models don't return a full set of class scores, but
    rather just a score for the predicted class. This is a helper function
    that approximates a full set of class scores by dividing the difference
    between 1.0 and the predicted class score among the remaning classes.

    :param scores: Set of class specific scores. Shape [D] where D is number
        of detections
    :type scores: torch.Tensor
    :param labels: Set of label indices corresponding to predicted class.
        Shape [D] where D is number of detections
    :type labels: torch.Tensor (ints)
    :param number_of_classes: Number of classes model predicts
    :type number_of_classes: int
    :return: A set of expanded scores, of shape [D, C], where C is number of
        classes
    :type: torch.Tensor
    """
    number_of_detections = scores.shape[0]

    expanded_scores = torch.ones(number_of_detections, number_of_classes + 1)

    for i, (score, label) in enumerate(zip(scores, labels)):

        residual = (1. - score.item()) / (number_of_classes)
        expanded_scores[i, :] *= residual
        expanded_scores[i, int(label.item())] = score

    return expanded_scores


def _wrap_image_model(model, examples, model_task, is_function,
                      number_of_classes: int = None,
                      classes: Union[list, np.array] = None,
                      device=Device.AUTO.value):
    """If needed, wraps the model or function in a common API.

    Wraps the model based on model task and prediction function contract.

    :param model: The model or function to evaluate on the examples.
    :type model: function or model to wrap
    :param examples: The model evaluation examples.
    :type examples: ml_wrappers.DatasetWrapper
    or numpy.ndarray or pandas.DataFrame or panads.Series
    or scipy.sparse.csr_matrix or shap.DenseData
    or torch.Tensor.
    :param model_task: Parameter to specify whether the model is an
    'image_classification' or another type of image model.
    :type model_task: str
    :param classes: optional parameter specifying a list of class names
    the dataset
    :type classes: list or np.ndarray
    :param number_of_classes: optional parameter specifying the
    number of classes in the dataset
    :type number_of_classes: int
    :param device: optional parameter specifying the device to move the model
        to.
    :type device: str
    :return: The function chosen from given model and chosen domain, or
    model wrapping the function and chosen domain.
    :rtype: (function, str) or (model, str)
    """
    device = _get_device(device)
    _wrapped_model = model
    if model_task == ModelTask.IMAGE_CLASSIFICATION:
        try:
            if isinstance(model, nn.Module):
                model = WrappedPytorchModel(model, image_to_tensor=True)
                if not isinstance(examples, DatasetWrapper):
                    examples = DatasetWrapper(examples)
                eval_function, eval_ml_domain = _eval_model(
                    model, examples, model_task)
                return (
                    WrappedClassificationModel(model, eval_function, examples),
                    eval_ml_domain,
                )
        except (NameError, AttributeError):
            module_logger.debug(
                'Could not import torch, required if using a pytorch model'
            )

        if _is_fastai_model(model):
            _wrapped_model = WrappedFastAIImageClassificationModel(model)
        elif hasattr(model, '_model_impl'):
            if str(type(model._model_impl.python_model)).endswith(
                ("azureml.automl.dnn.vision.common.mlflow."
                    "mlflow_model_wrapper.MLFlowImagesModelWrapper'>")
            ):
                _wrapped_model = WrappedMlflowAutomlImagesClassificationModel(
                    model)
        elif _is_transformers_pipeline(model) or _is_callable_pipeline(model):
            _wrapped_model = WrappedTransformerImageClassificationModel(model)
    elif model_task == ModelTask.MULTILABEL_IMAGE_CLASSIFICATION:
        if _is_fastai_model(model):
            _wrapped_model = WrappedFastAIImageClassificationModel(
                model, multilabel=True
            )
    elif model_task == ModelTask.OBJECT_DETECTION:
        if hasattr(model, '_model_impl'):
            if str(type(model._model_impl.python_model)).endswith(
                ("azureml.automl.dnn.vision.common.mlflow."
                    "mlflow_model_wrapper.MLFlowImagesModelWrapper'>")
            ):
                _wrapped_model = WrappedMlflowAutomlObjectDetectionModel(
                    model, classes)
        elif _is_transformers_pipeline(model) or _is_callable_pipeline(model):
            _wrapped_model = WrappedObjectDetectionModel(
                model, number_of_classes, device)
    return _wrapped_model, model_task


def _get_device(device: str) -> str:
    """Sets the device to run computations on to the desired value.

    If device were set to "auto", then the desired device will be gpu (CUDA)
    if available. Otherwise, the device should be set to cpu.

    :param device: parameter specifying the device to move the model
        to.
    :type device: str
    :return: selected device to run computations on
    :rtype: str
    """
    if (device in [member.value for member in Device] or type(device) is int or device.isdigit() or device is None):
        if device == Device.AUTO.value:
            if torch.cuda.is_available():
                return Device.CUDA.value
            else:
                return Device.CPU.value
        return device
    elif COLON in device:
        split_vals = device.split(COLON)
        if len(split_vals) == 2:
            return COLON.join([_get_device(
                split_val) for split_val in split_vals])
    raise ValueError("Selected device is invalid")


class WrappedTransformerImageClassificationModel(object):
    """A class for wrapping a Transformers model in the scikit-learn style."""

    def __init__(self, model):
        """Initialize the WrappedTransformerImageClassificationModel."""
        self._model = model

    def predict(self, dataset):
        """Predict the output using the wrapped Transformers model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted values.
        :rtype: numpy.ndarray
        """
        return np.argmax(self._model(dataset), axis=1)

    def predict_proba(self, dataset):
        """Predict the output probability using the Transformers model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted probabilities.
        :rtype: numpy.ndarray
        """
        return self._model(dataset)


class WrappedFastAIImageClassificationModel(object):
    """A class for wrapping a FastAI model in the scikit-learn style."""

    def __init__(self, model, multilabel=False):
        """Initialize the WrappedFastAIImageClassificationModel.

        :param model: The model to wrap.
        :type model: fastai.learner.Learner
        :param multilabel: Whether the model is a multilabel model.
        :type multilabel: bool
        """
        self._model = model
        self._multilabel = multilabel

    def _fastai_predict(self, dataset, index):
        """Predict the output using the wrapped FastAI model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :param index: The index into the predicted data.
            Index 1 is for the predicted class and index
            2 is for the predicted probability.
        :type index: int
        :return: The predicted data.
        :rtype: numpy.ndarray
        """
        # Note predict for single image requires 3d instead of 4d array
        if len(dataset.shape) == 4:
            predictions = []
            for row in dataset:
                predictions.append(self._fastai_predict(row, index))
            predictions = np.array(predictions)
            if index == 1 and not self._multilabel:
                predictions = predictions.flatten()
            return predictions
        else:
            predictions = np.array(self._model.predict(dataset)[index])
            if len(predictions.shape) == 0:
                predictions = predictions.reshape(1)
            if index == 1:
                is_boolean = predictions.dtype == bool
                if is_boolean:
                    predictions = predictions.astype(int)
            return predictions

    def predict(self, dataset):
        """Predict the output value using the wrapped FastAI model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted values.
        :rtype: numpy.ndarray
        """
        return np.array(self._fastai_predict(dataset, 1))

    def predict_proba(self, dataset):
        """Predict the output probability using the FastAI model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted probabilities.
        :rtype: numpy.ndarray
        """
        return self._fastai_predict(dataset, 2)


class WrappedMlflowAutomlImagesClassificationModel:
    """A class for wrapping an AutoML for images MLflow classification
        model in the scikit-learn style."""

    def __init__(self, model: PyFuncModel) -> None:
        """Initialize the WrappedMlflowAutomlImagesClassificationModel.

        :param model: mlflow model
        :type model: mlflow.pyfunc.PyFuncModel
        """
        self._model = model

    def _mlflow_predict(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """Perform the inference using the wrapped MLflow model.

        :param dataset: The dataset to predict on.
        :type dataset: pandas.DataFrame
        :return: The predicted data.
        :rtype: pandas.DataFrame
        """
        predictions = self._model.predict(dataset)
        return predictions

    def predict(self, dataset: pd.DataFrame) -> np.ndarray:
        """Predict the output value using the wrapped MLflow model.

        :param dataset: The dataset to predict on.
        :type dataset: pandas.DataFrame
        :return: The predicted values.
        :rtype: numpy.ndarray
        """
        predictions = self._mlflow_predict(dataset)
        return predictions.loc[:, 'probs'].map(lambda x: np.argmax(x)).values

    def predict_proba(self, dataset: pd.DataFrame) -> np.ndarray:
        """Predict the output probability using the MLflow model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: pandas.DataFrame
        :return: The predicted probabilities.
        :rtype: numpy.ndarray
        """
        predictions = self._mlflow_predict(dataset)
        return np.stack(predictions.probs.values)


class WrappedObjectDetectionModel:
    """A class for wrapping a object detection model in the scikit-learn
        style."""

    def __init__(self,
                 model: Any,
                 number_of_classes: int,
                 device=Device.AUTO.value) -> None:
        """Initialize the WrappedObjectDetectionModel with the model
            and evaluation function.

        :param model: mlflow model
        :type model: Any
        :param number_of_classes: number of distinct labels
        :type number_of_classes: int
        :param device: optional parameter specifying the device to move the
            model to. If not specified, then cpu is the default
        :type device: str
        """
        self._device = torch.device(_get_device(device))
        model.eval()
        model.to(self._device)

        self._model = model
        self._number_of_classes = number_of_classes

    def predict(self, x, iou_threshold: float = 0.5, score_threshold: float = 0.5):
        """Create a list of detection records from the image predictions.

        :param x: Tensor of the image
        :type x: torch.Tensor
        :return: Baseline detections to get saliency maps for
        :rtype: numpy array of Detection Records

        Example Label (y) representation for a cohort of 2 images:

        [

            [
            [object_1, x1, y1, b1, h1, (optional) confidence_score],
            [object_2, x2, y2, b2, h2, (optional) confidence_score],
            [object_1, x3, y3, b3, h3, (optional) confidence_score]
        ],

        [
            [object_1, x4, y4, b4, h4, (optional) confidence_score],
            [object_2, x5, y5, b5, h5, (optional) confidence_score]
        ]

        ]

        """
        detections = []
        for image in x:
            if type(image) is Tensor:
                raw_detections = self._model(
                    image.to(self._device).unsqueeze(0))
            else:
                raw_detections = self._model(
                    T.ToTensor()(image).to(self._device).unsqueeze(0))

            for raw_detection in raw_detections:
                raw_detection = _apply_nms(raw_detection, .5)
                raw_detection = _filter_score(raw_detection)
                image_predictions = torch.cat((raw_detection[LABELS]
                                               .unsqueeze(1),
                                               raw_detection[BOXES],
                                               raw_detection[SCORES]
                                               .unsqueeze(1)), dim=1)

                detections.append(image_predictions.detach().cpu().numpy()
                                  .tolist())
        try:
            return np.array(detections)
        except ValueError:
            return detections

    def predict_proba(self, dataset, iou_threshold=0.1):
        """Predict the output probability using the wrapped model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        :param iou_threshold: amount of acceptable error.
            objects with error scores higher than the threshold will be removed
        :type iou_threshold: float
        """
        predictions = self.predict(dataset, iou_threshold)
        prob_scores = [[pred[-1] for pred in image_prediction]
                       for image_prediction in predictions]
        return prob_scores


class WrappedMlflowAutomlObjectDetectionModel:
    """A class for wrapping an AutoML for images MLflow object detection model
        in the scikit-learn style."""

    def __init__(self, model: PyFuncModel,
                 classes: Union[list, np.array]) -> None:
        """Initialize the WrappedMlflowAutomlObjectDetectionModel.

        :param model: mlflow model
        :type model: mlflow.pyfunc.PyFuncModel
        :param classes: list of class names
        :type classes: list or np.array
        """

        self._model = model
        self._classes = classes
        try:
            if type(classes[0] == str):
                self._label_dict = {label: (i + 1)
                                    for i, label in enumerate(classes)}
        except KeyError:
            raise KeyError("classes parameter not a list of class labels")

    def _mlflow_predict(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """Perform the inference using the wrapped MLflow model.

        :param dataset: The dataset to predict on.
        :type dataset: pandas.DataFrame
        :return: The predicted data.
        :rtype: pandas.DataFrame
        """

        predictions = self._model.predict(dataset)
        return predictions

    def predict(self, dataset: pd.DataFrame, iou_threshold: float = 0.5,
                score_threshold: float = 0.5):
        """Create a list of detection records from the image predictions.

        Below is example Label (y) representation for a cohort of 2 images,
        with 3 objects detected for the first image and 1 for the second image.

        [
         [
            [class, topX, topY, bottomX, bottomY, (optional) confidence_score],
            [class, topX, topY, bottomX, bottomY, (optional) confidence_score],
            [class, topX, topY, bottomX, bottomY, (optional) confidence_score],
         ],
         [
            [class, topX, topY, bottomX, bottomY, (optional) confidence_score],
         ]

        ]

        :param dataset: The dataset to predict on.
        :type dataset: pandas.DataFrame
        :param iou_threshold: Intersection-over-Union (IoU) threshold for NMS (or
                           the amount of acceptable error). Objects with error
                           cores higher than the threshold will be removed.
        :type iou_threshold: float
        :param score_threshold: Threshold to filter detections based on
                            predicted confidence scores.
        :type score_threshold: float
        :return: Final detections from the object detector
        :rtype: numpy array of Detection Records
        """
        image_sizes = dataset['image_size']

        dataset = dataset.drop(['image_size'], axis=1)

        predictions = self._mlflow_predict(dataset)
        if not len(predictions['boxes']) == len(image_sizes):
            raise ValueError("Internal Error: Number of predictions "
                             "does not match number of images")

        detections = []
        for image_detections, img_size in \
                zip(predictions['boxes'], image_sizes):

            # process automl mlflow detection outputs
            raw_detections = _process_automl_detections_to_raw_detections(
                image_detections, self._label_dict, img_size)

            raw_detections = _apply_nms(raw_detections, iou_threshold)
            raw_detections = _filter_score(raw_detections, score_threshold)

            image_predictions = torch.cat((
                raw_detections[LABELS].unsqueeze(1),
                raw_detections[BOXES],
                raw_detections[SCORES].unsqueeze(1)
            ),
                dim=1
            )

            detections.append(
                image_predictions.detach().cpu().numpy().tolist())

        try:
            return np.array(detections)
        except ValueError:
            return detections

    def predict_proba(self, dataset: pd.DataFrame,
                      iou_threshold=0.1) -> np.ndarray:
        """Predict the output probability using the MLflow model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: pandas.DataFrame
        :param iou_threshold: amount of acceptable error.
            objects with error scores higher than the threshold will be removed
        :type iou_threshold: float
        :return: The predicted probabilities.
        :rtype: numpy.ndarray
        """

        predictions = self.predict(dataset, iou_threshold=iou_threshold)
        prob_scores = [[pred[-1] for pred in image_prediction]
                       for image_prediction in predictions]
        return prob_scores


class PytorchDRiseWrapper(GeneralObjectDetectionModelWrapper):
    """Wraps a PytorchFasterRCNN model with a predict API function.

    To be compatible with the D-RISE explainability method,
    all models must be wrapped to have the same output and input class and a
    predict function for object detection. This wrapper is customized for the
    FasterRCNN model from Pytorch, and can also be used with the RetinaNet or
    any other models with the same output class.
    """

    def __init__(self, model, number_of_classes: int,
                 device=Device.AUTO.value,
                 transforms=None,
                 iou_threshold=None,
                 score_threshold=None):
        """Initialize the PytorchDRiseWrapper.

        :param model: Object detection model
        :type model: PytorchFasterRCNN model
        :param number_of_classes: Number of classes the model is predicting
        :type number_of_classes: int
        :param device: Optional parameter specifying the device to move the
            model to. If not specified, then cpu is the default
        :type device: str
        :param transforms: Optional parameter specifying the transforms to
            apply to the image before passing it to the model
        :type transforms: torchvision.transforms
        :param iou_threshold: Optional parameter specifying the iou_threshold
            for nms. If not specified, the iou_threshold on the predict
            method is used.
        :type iou_threshold: float
        :param score_threshold: Optional parameter specifying the
            score_threshold for filtering detections. If not
            specified, the score_threshold on the predict method is
            used.
        :type score_threshold: float
        """
        self.transforms = transforms
        self._device = torch.device(_get_device(device))
        model.to(self._device)
        model.eval()

        self._model = model
        self._number_of_classes = number_of_classes
        self._iou_threshold = iou_threshold
        self._score_threshold = score_threshold

    def predict(self, x: Tensor, iou_threshold: float = 0.5,
                score_threshold: float = 0.5):
        """Create a list of detection records from the image predictions.

        :param x: Tensor of the image
        :type x: torch.Tensor
        :param iou_threshold: Intersection-over-Union (IoU) threshold for NMS (or
            the amount of acceptable error). Objects with error
            scores higher than the threshold will be removed.
        :type iou_threshold: float
        :param score_threshold: Threshold to filter detections based on
                            predicted confidence scores.
        :type score_threshold: float
        :return: Baseline detections to get saliency maps for
        :rtype: List of Detection Records
        """
        if self._iou_threshold is not None:
            iou_threshold = self._iou_threshold
        if self._score_threshold is not None:
            score_threshold = self._score_threshold
        with torch.no_grad():
            raw_detections = self._model(x)

            detections = []
            for raw_detection in raw_detections:
                raw_detection = _apply_nms(raw_detection, iou_threshold)

                # Note that FasterRCNN doesn't return a score for each
                # class, only the predicted class. DRISE requires a
                # score for each class.
                # We approximate the score for each class
                # by dividing (class score) evenly among the other classes.

                raw_detection = _filter_score(raw_detection, score_threshold)
                expanded_class_scores = expand_class_scores(
                    raw_detection[SCORES],
                    raw_detection[LABELS],
                    self._number_of_classes)

                detections.append(
                    od_common.DetectionRecord(
                        bounding_boxes=raw_detection[BOXES],
                        class_scores=expanded_class_scores,
                        objectness_scores=torch.tensor(
                            [1.0] * raw_detection[BOXES].shape[0]),
                    )
                )

            return detections


class MLflowDRiseWrapper():
    """Wraps a Mlflow model with a predict API function.

    To be compatible with the D-RISE explainability method,
    all models must be wrapped to have the same output and input class and a
    predict function for object detection. This wrapper is customized for the
    FasterRCNN model from AutoML. Unlike the Pytorch wrapper, this wrapper
    does not inherit from GeneralObjectDetectionModelWrapper as this super
    class requires predict to take a tensor input.
    """

    def __init__(self, model: PyFuncModel,
                 classes: Union[list, np.ndarray]) -> None:
        """Initialize the MLflowDRiseWrapper.

        :param model: mlflow model
        :type model: mlflow.pyfunc.PyFuncModel
        :param number_of_classes: Number of classes the model is predicting
        :type number_of_classes: int
        """

        self._model = model
        self._classes = classes
        self._number_of_classes = len(classes)
        self._label_dict = {label: (i + 1) for i, label in enumerate(classes)}

    def _mlflow_predict(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """Perform the inference using the wrapped MLflow model.

        :param dataset: The dataset to predict on.
        :type dataset: pandas.DataFrame
        :return: The predicted data.
        :rtype: pandas.DataFrame
        """
        predictions = self._model.predict(dataset)
        return predictions

    def predict(self, dataset: pd.DataFrame, iou_threshold: float = 0.25,
                score_threshold: float = 0.5):
        """Predict the output value using the wrapped MLflow model.

        :param dataset: The dataset to predict on.
        :type dataset: pandas.DataFrame
        :param iou_threshold: Intersection-over-Union (IoU) threshold for NMS (or
            the amount of acceptable error). Objects with error
            scores higher than the threshold will be removed.
        :type iou_threshold: float
        :param score_threshold: Threshold to filter detections based on
                            predicted confidence scores.
        :type score_threshold: float
        :return: The predicted values.
        :rtype: numpy.ndarray
        """
        image_sizes = dataset['image_size']

        dataset = dataset.drop(['image_size'], axis=1)

        predictions = self._mlflow_predict(dataset)
        if not len(predictions['boxes']) == len(image_sizes):
            raise ValueError("Internal Error: Number of predictions "
                             "does not match number of images")

        if not len(predictions['boxes']) == 1:
            raise ValueError(
                "Currently, only 1 image can be passed to predict")

        detections = []
        for image_detections, img_size in \
                zip(predictions['boxes'], image_sizes):

            raw_detections = _process_automl_detections_to_raw_detections(
                image_detections, self._label_dict, img_size)

            # TODO: check if this is needed
            # No detections found - most likely in masked image
            if raw_detections[BOXES].nelement() == 0:
                detections.append(None)
                continue

            raw_detections = _apply_nms(raw_detections, iou_threshold)
            raw_detections = _filter_score(raw_detections, score_threshold)

            # Note that FasterRCNN doesn't return a score for each class, only
            # the predicted class. DRISE requires a score for each class.
            # We approximate the score for each class
            # by dividing (class score) evenly among the other classes.

            expanded_class_scores = expand_class_scores(
                raw_detections[SCORES],
                raw_detections[LABELS],
                self._number_of_classes)

            detections.append(
                od_common.DetectionRecord(
                    bounding_boxes=raw_detections[BOXES],
                    class_scores=expanded_class_scores,
                    objectness_scores=torch.tensor(
                        [1.0] * raw_detections[BOXES].shape[0]),
                ))

        return detections



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines model wrappers and utilities for pytorch models."""

import logging

import numpy as np
import pandas as pd

module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


try:
    import torch
except ImportError:
    module_logger.debug('Could not import torch, required if using a PyTorch model')

try:
    from torchvision.transforms import ToTensor
except ImportError:
    module_logger.debug('Could not import torchvision, required if using'
                        ' a vision PyTorch model')


class WrappedPytorchModel(object):
    """A class for wrapping a PyTorch model.

    Note at time of initialization, since we don't have
    access to the dataset, we can't infer if this is for
    classification or regression case.  Hence, we add
    the predict_classes method for classification, and keep
    predict for either outputting values in regression or
    probabilities in classification.
    """

    def __init__(self, model, image_to_tensor=False):
        """Initialize the WrappedPytorchModel with the model and evaluation function.

        :param model: The PyTorch model to wrap.
        :type model: torch.nn.Module
        :param image_to_tensor: Whether to convert the image to tensor.
        :type image_to_tensor: bool
        """
        self._model = model
        # Set eval automatically for user for batchnorm and dropout layers
        self._model.eval()
        self._image_to_tensor = image_to_tensor

    def _convert_to_tensor(self, dataset):
        """Convert the dataset to a pytorch tensor.

        For image datasets, we use ToTensor from torchvision,
        which moves channel to the first dimension and for
        2D images adds a third dimension.

        :param dataset: The dataset to convert.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The converted dataset.
        :rtype: torch.Tensor
        """
        # Convert the data to pytorch Variable
        if isinstance(dataset, pd.DataFrame):
            if self._image_to_tensor:
                dataset = dataset.iloc[0]
            dataset = dataset.values
            # If represented as a list of arrays,
            # convert to a 3D array instead of array
            # of 2D arrays
            if len(dataset.shape) == 1:
                if self._image_to_tensor and len(dataset[0].shape) == 2:
                    # add channel to end of image if 2D grayscale
                    for i in range(dataset.shape[0]):
                        dataset[i] = np.expand_dims(dataset[i], axis=2)
                dataset = np.stack(dataset)
        # If not already tensor, convert
        if not isinstance(dataset, torch.Tensor):
            if self._image_to_tensor:
                # For torchvision images, can only convert one
                # image at a time
                # Note pytorch wrapper expects extra dimension for rows
                # to be expanded in evaluator for image case,
                # otherwise this code won't work for a single
                # image input to predict call
                rows = []
                for row in range(dataset.shape[0]):
                    instance = dataset[row]
                    if not isinstance(instance, torch.Tensor):
                        instance = ToTensor()(instance)
                    rows.append(instance)
                dataset = torch.stack(rows)
            else:
                dataset = torch.Tensor(dataset)
        return dataset

    def predict(self, dataset):
        """Predict the output using the wrapped PyTorch model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The prediction results.
        :rtype: numpy.ndarray
        """
        wrapped_dataset = self._convert_to_tensor(dataset)
        with torch.no_grad():
            result = self._model(wrapped_dataset).numpy()
        # Reshape to 2D if output is 1D and input has one row
        if len(dataset.shape) == 1:
            result = result.reshape(1, -1)
        return result

    def predict_classes(self, dataset):
        """Predict the class using the wrapped PyTorch model.

        :param dataset: The dataset to predict on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted classes.
        :rtype: numpy.ndarray
        """
        wrapped_dataset = self._convert_to_tensor(dataset)
        with torch.no_grad():
            result = self._model(wrapped_dataset)
        result_len = len(result.shape)
        if result_len == 1 or (result_len > 1 and result.shape[1] == 1):
            result = np.where(result.numpy() > 0.5, 1, 0)
        else:
            result = torch.max(result, 1)[1].numpy()
        # Reshape to 2D if output is 1D and input has one row
        if len(dataset.shape) == 1:
            result = result.reshape(1, -1)
        return result

    def predict_proba(self, dataset):
        """Predict the output probability using the wrapped PyTorch model.

        :param dataset: The dataset to predict_proba on.
        :type dataset: ml_wrappers.DatasetWrapper
        :return: The predicted probabilities.
        :rtype: numpy.ndarray
        """
        return self.predict(dataset)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a common dataset wrapper and common functions for data manipulation."""

from .dataset_wrapper import DatasetWrapper
from .timestamp_featurizer import CustomTimestampFeaturizer

__all__ = ['CustomTimestampFeaturizer', 'DatasetWrapper']
